/* //device/samples/SampleCode/src/com/android/samples/app/RemoteServiceInterface.java
**
** Copyright 2007, The Android Open Source Project
**
** Licensed under the Apache License, Version 2.0 (the "License"); 
** you may not use this file except in compliance with the License. 
** You may obtain a copy of the License at 
**
**     http://www.apache.org/licenses/LICENSE-2.0 
**
** Unless required by applicable law or agreed to in writing, software
** distributed under the License is distributed on an "AS IS" BASIS, 
** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
** See the License for the specific language governing permissions and 
** limitations under the License.
*/

package com.baidu.music.logic.service;

import com.baidu.music.logic.service.IMusicPlayCallback;
import com.baidu.music.logic.service.RadioChannel;

interface IMusicPlayService
{
	void playCurrent();
	void playNext(int cause);
	void playPrevious();
	void playAtPosition(int position);
	void playFromWidget();
    void playRadioChannel(in RadioChannel rc);
    void playNextFromWidget();
    void playPreviousFromWidget();
	void play();
	void pause();
	void pauseByCarMode();
	void stop();
	void  setServiceState(boolean status);
	
	void seek(long pos);
	long getSongId();
	long getTingUid();
	long getMusicInfoId();
	int getPlayListPosition();
	
	void changeNextMode();
	int getPlayMode();
	void setDefaultPlayMode(int playMode);
	 void setNormalMode();

	void setRepeatOneMode();

	void setRepeatAllMode();

	void setShuffleMode();
	String getArtistName();

	String getAudioName();

	String getAlbumName();

	//long getTotalPlayCount();
		
	String getShowUrl();
	String getResourceType();
	
	boolean isHighQuality() ;
	
	boolean isPlaying() ;

	boolean isBuffering();

	boolean isEmptyPlayList();

	boolean isStopped();
	
	boolean isPaused();
	
	boolean isIdle();
	
	long position();
	
	long duration();
	
	long getDownloadProgress();
	
	
	
	String[] getLyricSentences();
	
	long[] getFromTimes();
	
	long[] getToTimes();
	  
	String getImagePath();
	String getImageLink();
	String getAlbumImage();
	  
	int getLyricState();
	int getAudioType();
	int getDownloadState();
	int getFavriteState();

	void registerCallback(IMusicPlayCallback cb);
	void unregisterCallback(IMusicPlayCallback cb);
	boolean isPlayRadio();
	boolean isPlayLocal();
	boolean isPlayScene();
	void notifyChange(String what);
         
	void useEQpreset(int preset);
	void setSurroundLevel(int level);
	void getSurroundLevelRange(out int[] range);
	void enableEffect(int nEffectType, boolean bEnable);
	void getFreqValueRange(out int[] range);
	void getFreqData(int datacount,out int[] data);
	void enableAllEffect(boolean enabled);
	void setEQBandLevel(int band, int level);
	void getEQBandLevelRange(out int[]range);
	int getEQNumberOfBands();

	
	String getLyricRawString();

	boolean canKtv();
	
	void playScene(int sceneId);
	int getCurrentSourceType();
 	String getCurrentSongPath();
    boolean hasPlayedSong();
    String getAudioVersion();
    void setAudioEffectCombine(String path);
    void resetAudioEffect();
    void setEQLevels(out int [] levels);
    void getEQLevels(out int [] levels);
    void getPresetEQLevels(int eq,out int [] levels);
    void setBassLevel(int level);
    int getCurrentSongType();
}

