package com.baidu.musicPlayer;

import java.io.IOException;

import com.baidu.util.learncore.LearnPlayer;
import com.baidu.utility.LogHelper;

import android.media.AudioFormat;
import android.media.AudioManager;
import android.media.AudioTrack;
import android.os.Handler;
import android.os.Handler.Callback;
import android.os.HandlerThread;
import android.os.Message;
import android.os.SystemClock;

public class AudioTrackPlayer {
    private static final String tag = "AudioTrackPlayer";
    private LogHelper logHelper = new LogHelper(tag);

    // for render
    private static final int RENDER_START = 1;
    private static final int RENDER_RUN = 2;
    private static final int RENDER_STOP = 3;

    private static final int RENDER_ONCE_BUFFER_SIZE = 8192;

    int mStreamtype = AudioManager.STREAM_MUSIC;

    private AudioTrack mAudioTrack;
    private LearnRenderHandlerThread mRenderThread;
    private Handler mRenderhandler;
    private int mBufferSize = 0;
    private int mMiniBufferDevice = 8192;
    private int mStartPos = 0;
    private int mStartPosTem = 0;
    // private int mSessionID = 0;
    private int mSampleRateInHz = 44100;
    private int mChannels = 2;
    private int mBitPerSample = 16;
    private boolean mFirstFrame = false;
    private long mAudioTrackSum = 0;
    private long mAudioTrackSumStart = 0;
    private int mBytePerSec = 0;
    private LearnPlayer mLearnPlayer = null;

    // private int mStartSamples = 0;

    private boolean mIsPlaying = false;

    private class LearnRenderHandlerThread extends HandlerThread implements Callback {
        private byte[] mData;

        // private boolean mIsSetPriority;

        public LearnRenderHandlerThread(String name) {
            super(name);
            mData = new byte[RENDER_ONCE_BUFFER_SIZE];
            // mIsSetPriority = false;
        }

        @Override
        public boolean handleMessage(Message msg) {
            // 设置线程优先级，保证资源分配。
            // android.os.Process.THREAD_PRIORITY_URGENT_AUDIO is too high,
            // some synchronized may be blocked
            // if(!mIsSetPriority) {
            // int nTid = android.os.Process.myTid();
            // int nOriPriorty = android.os.Process.getThreadPriority(nTid);
            // // 设置线程优先级，保证资源分配。
            // android.os.Process
            // .setThreadPriority(android.os.Process.THREAD_PRIORITY_URGENT_AUDIO);
            //
            // int nCurPriorty = android.os.Process.getThreadPriority(nTid);
            //
            // logHelper.infoShow("RenderHandlerThreadPriority nTid "+ nTid + "nOriPriorty " + nOriPriorty +
            // "nCurPriorty" + nCurPriorty);
            //
            // mIsSetPriority = true;
            // }

            // logHelper.infoShow( msg.toString());
            if (msg.what == RENDER_RUN) {
                // logHelper.infoShow("begin getRenderBuffer");

                int sizeInBytes =
                        mLearnPlayer.getRenderBuffer(mData, mBufferSize < mData.length ? mBufferSize : mData.length);
                // logHelper.infoShow("end getRenderBuffer sizeInBytes:" + sizeInBytes);

                if (sizeInBytes > 0) {
                    // logHelper.infoShow( "AudioTrack write begin.");
                    synchronized (AudioTrackPlayer.this) {
                        if ((mAudioTrack != null) && (sizeInBytes <= 10240)) {
                            mAudioTrack.write(mData, 0, sizeInBytes);
                            mAudioTrackSum += sizeInBytes; // ZTE blocks when
                                                           // getPlaybackHeadPosition
                        }
                        if (mAudioTrack == null) {
                            logHelper.infoShow("mAudioTrack = null, sizeInBytes=%d" + sizeInBytes);
                        }
                    }
                    // logHelper.infoShow( "AudioTrack write end. sizeInBytes:" + sizeInBytes);

                    Handler Renderhandler = mRenderhandler;
                    if (Renderhandler != null) {
                        Renderhandler.sendEmptyMessage(RENDER_RUN);
                    }

                    if (mFirstFrame) {
                        mFirstFrame = false;
                        // MEDIA_INFO = 200
                        // MEDIA_INFO_RENDER_START = 900
                        logHelper.infoShow("AudioTrack first write");
                        mLearnPlayer.sendMessage(200, 900, 0, null);
                    }

                } else {
                    // Wait 1ms.
                    Handler Renderhandler = mRenderhandler;
                    // logHelper.infoShow( "render sleep 1ms ");
                    if (Renderhandler != null) {
                        Renderhandler.sendEmptyMessageAtTime(RENDER_RUN, SystemClock.uptimeMillis() + 1);
                        // Renderhandler.sendEmptyMessageDelayed(RENDER_RUN,5);
                    }
                    // try {
                    // Thread.sleep(10);
                    // } catch (InterruptedException e) {
                    // e.printStackTrace();
                    // }
                    // mRenderhandler.sendEmptyMessage(RENDER_RUN);
                }
            } else if (msg.what == RENDER_START) {
                logHelper.infoShow("render RENDER_START start");
                synchronized (AudioTrackPlayer.this) {
                    Handler Renderhandler = mRenderhandler;
                    if (Renderhandler != null) {
                        Renderhandler.sendEmptyMessage(RENDER_RUN);
                    }
                }
                logHelper.infoShow("render RENDER_START end");
            } else if (msg.what == RENDER_STOP) {
                logHelper.infoShow("render RENDER_STOP start");
                synchronized (AudioTrackPlayer.this) {
                    Handler Renderhandler = mRenderhandler;
                    if (Renderhandler != null) {
                        Renderhandler.removeMessages(RENDER_RUN);
                    }
                }
                logHelper.infoShow("render RENDER_STOP end");
            }
            return true;
        }
    }

    public AudioTrackPlayer() {
        // Create render thread.
        mRenderThread = new LearnRenderHandlerThread("render");
        mRenderThread.start();
        mRenderhandler = new Handler(mRenderThread.getLooper(), mRenderThread);
    }

    public AudioTrackPlayer(LearnPlayer learnPlayer) {
        mRenderThread = new LearnRenderHandlerThread("render");
        mRenderThread.start();
        mRenderhandler = new Handler(mRenderThread.getLooper(), mRenderThread);
        mLearnPlayer = learnPlayer;
        // TODO Auto-generated constructor stub
    }

    /**
     * Prepares the player for playback, synchronously.
     * 
     * After setting the datasource and the display surface, you need to either call prepare() or prepareAsync(). For
     * files, it is OK to call prepare(), which blocks until AudioPlayer is ready for playback.
     * 
     * @throws IllegalStateException if it is called in an invalid state
     */
    public void prepare() throws IOException, IllegalStateException {
        mStartPos = 0;
        // mStartSamples = 0;
        mAudioTrackSum = 0;
        mAudioTrackSumStart = 0;

        mBytePerSec = mSampleRateInHz * mChannels * mBitPerSample / 8;

        mIsPlaying = false;
        mFirstFrame = true;
    }

    public void setSampleRate(int nSamplerate) {
        mSampleRateInHz = nSamplerate;
    }

    public void setChannels(int channel) {
        mChannels = channel;
    }

    public void setBitsPerSample(int bitpersample) {
        mBitPerSample = bitpersample;
    }

    /**
     * Starts or resumes playback. If playback had previously been paused, playback will continue from where it was
     * paused. If playback had been stopped, or never started before, playback will start at the beginning.
     * 
     * @throws IllegalStateException if it is called in an invalid state
     */
    public void start() throws IllegalStateException {
        logHelper.infoShow("start start ");
        int channelConfig = mChannels;
        int audioFormat = mBitPerSample;
        int bytepersec = mBytePerSec;

        if (mChannels == 1) {
            channelConfig = AudioFormat.CHANNEL_OUT_MONO;
        } else {
            channelConfig = AudioFormat.CHANNEL_OUT_STEREO;
        }
        // Only support 16bit.
        audioFormat = AudioFormat.ENCODING_PCM_16BIT;

        int minBufSize = AudioTrack.getMinBufferSize(mSampleRateInHz, channelConfig, audioFormat);
        if (minBufSize <= 0)
            throw new IllegalStateException("minBufSize less than zero");
        logHelper.infoShow("minBufSize = " + minBufSize);

        mMiniBufferDevice = minBufSize;
        AudioTrack audioTrack = mAudioTrack;
        if (audioTrack == null) {
            synchronized (this) {
                // max buffer is 0.5 sec

                // minBufSize = minBufSize * 2;
                // int maxBufSize = minBufSize > bytepersec / 2 ? minBufSize
                // : bytepersec / 2;
                // minBufSize = minBufSize * 8 > maxBufSize ? maxBufSize
                // : minBufSize * 8;
                logHelper.infoShow("bytepersec = " + bytepersec);
                minBufSize = minBufSize * 4; // 4 times
                if (minBufSize > (bytepersec / 2)) {
                    minBufSize = bytepersec / 2;
                }
                minBufSize = (minBufSize / 1024) * 1024;

                if (minBufSize < RENDER_ONCE_BUFFER_SIZE) {
                    minBufSize = RENDER_ONCE_BUFFER_SIZE;
                }
                logHelper.infoShow("minBufSize = " + minBufSize);

                if (mAudioTrack == null) // if construct with sessionid not
                                         // available, use old constructor
                {
                    mAudioTrack =
                            new AudioTrack(mStreamtype, mSampleRateInHz, channelConfig, audioFormat, minBufSize,
                                    AudioTrack.MODE_STREAM);
                }

                mBufferSize = minBufSize;
                audioTrack = mAudioTrack;
                // mStartSamples = 0;
                mAudioTrackSumStart = 0;
            }
        }
        audioTrack.play();

        if (mRenderhandler != null) {
            mRenderhandler.removeMessages(RENDER_START);
            mRenderhandler.sendEmptyMessage(RENDER_START);
        }

        mIsPlaying = true;
        logHelper.infoShow("start end ");
    }

    /**
     * Stops playback after playback has been stopped or paused.
     * 
     * @throws IllegalStateException if the internal player engine has not been initialized.
     */
    public void stop() throws IllegalStateException {
        logHelper.infoShow("stop start ");
        releaseAudiotrack();

        mIsPlaying = false;
        mFirstFrame = false;
        logHelper.infoShow("stop end ");
    }

    /**
     * Pauses playback. Call start() to resume.
     * 
     * @throws IllegalStateException if the internal player engine has not been initialized.
     */
    public void pause() throws IllegalStateException {
        logHelper.infoShow("pause start ");
        if (mRenderhandler != null) {
            mRenderhandler.removeMessages(RENDER_STOP);
            mRenderhandler.sendEmptyMessage(RENDER_STOP);
        }
        AudioTrack audioTrack = mAudioTrack;
        if (audioTrack != null) {
            audioTrack.pause();
        }

        mIsPlaying = false;
        logHelper.infoShow("pause end ");
    }

    /**
     * Checks whether the AudioPlayer is playing.
     * 
     * @return true if currently playing, false otherwise
     */

    public boolean isPlaying() {
        boolean isPlaying = false;
        if (mAudioTrack != null) {
            isPlaying = mIsPlaying;
        }
        return isPlaying;
    }

    public int getDeviceMinbufferMS() {
        if (mBytePerSec == 0) {
            return 100;
        }
        return (int) (((long) mMiniBufferDevice) * 1000 / mBytePerSec);
    }

    public int getMiniBufferMS() {
        if (mBytePerSec == 0) {
            return 400;
        }
        return (int) (((long) mBufferSize) * 1000 / mBytePerSec);
    }

    /**
     * Seeks to specified time position.
     * 
     * @param msec the offset in milliseconds from the start to seek to
     * @throws IllegalStateException if the internal player engine has not been initialized
     */
    public void seekTo(int msec) throws IllegalStateException {
        logHelper.infoShow("seekTo start " + msec);
        if (msec >= 0) {
            mStartPosTem = msec;
        }
        logHelper.infoShow("seekTo end mStartPosTem " + mStartPosTem);
    }

    /**
     * Gets the current playback position.
     * 
     * @return the current position in milliseconds
     */
    public int getCurrentPosition() {
        int position = 0;
        AudioTrack audioTrack = mAudioTrack;
        try {
            if (audioTrack != null) {
                int playRate = mSampleRateInHz;
                if (playRate <= 0)
                    playRate = 44100;

                // position = mStartPos
                // + (int) (((long) (audioTrack.getPlaybackHeadPosition() - mStartSamples) * 1000) / playRate);

                // logHelper.debugShow("getCurrentPosition:" + position + "mStartPos: " + mStartPos);

                long lAudioTrackByte = mAudioTrackSum - mAudioTrackSumStart;
                lAudioTrackByte -= mBufferSize - 15; // 15ms is delay between audioTrack.getPlaybackHeadPosition() and
                                                     // mBufferSize
                if (lAudioTrackByte < 0)
                    lAudioTrackByte = 0;

                if (mBytePerSec != 0) {
                    position = mStartPos + (int) ((lAudioTrackByte * 1000) / mBytePerSec);
                }
                // if(position > 0) position += 1000;
                // logHelper.debugShow("getCurrentPosition position:" + position+ "mStartPos:"+mStartPos
                // +"mAudioTrackSum " + mAudioTrackSum + "mAudioTrackSumStart " + mAudioTrackSumStart);
            }
        } catch (Exception ex) {
            logHelper.errorShow(ex.toString());
        }
        if (position <= 0)
            position = 1;
        return position;
    }

    /**
     * Releases resources associated with this AudioPlayer object. It is considered good practice to call this method
     * when you're done using the AudioPlayer.
     */
    public void release() {
        logHelper.infoShow("release start ");
        releaseAudiotrack();
        if (mRenderThread != null) {
            mRenderThread.getLooper().quit();
        }
        mRenderThread = null;
        mRenderhandler = null;

        logHelper.infoShow("release end ");
    }

    private void releaseAudiotrack() {
        if (mRenderhandler != null) {
            // logHelper.infoShow( "begin removeMessages RENDER_STOP!");
            mRenderhandler.removeMessages(RENDER_STOP);
            // logHelper.infoShow( "begin sendEmptyMessage RENDER_STOP!");
            mRenderhandler.sendEmptyMessage(RENDER_STOP);
        }
        synchronized (this) {
            AudioTrack audioTrack = mAudioTrack;
            try {
                if (audioTrack != null) {
                    audioTrack.pause();
                    audioTrack.flush();
                    audioTrack.stop();
                    audioTrack.release();
                }
            } catch (Exception e) {
                e.printStackTrace();
            }

            mAudioTrack = null;
        }
    }

    /**
     * Resets the AudioPlayer to its uninitialized state. After calling this method, you will have to initialize it
     * again by setting the data source and calling prepare().
     */
    public void reset() {
        logHelper.infoShow("reset start ");
        releaseAudiotrack();

        mStartPos = 0;
        mAudioTrackSumStart = 0;

        // mStartSamples = 0;

        mIsPlaying = false;
        mFirstFrame = false;
        logHelper.infoShow("reset end ");
    }

    /**
     * Sets the audio stream type for this AudioPlayer. See {@link AudioManager} for a list of stream types. Must call
     * this method before prepare() or prepareAsync() in order for the target stream type to become effective
     * thereafter.
     * 
     * @param streamtype the audio stream type
     * @see com.baidu.util.audiocore.AudioManager
     */
    public void setAudioStreamType(int streamtype) {
        mStreamtype = streamtype;
    }

    public void OnseekComplete(boolean bSuc) {
        if (bSuc) {
            {
                AudioTrack auTrack = mAudioTrack;
                try {
                    if (auTrack != null) {
                        auTrack.flush();
                        // mStartSamples =
                        // audioTrack.getPlaybackHeadPosition();
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                }

                mAudioTrackSumStart = mAudioTrackSum;
            }
            mStartPos = mStartPosTem;
            // logHelper.debugShow("OnseekComplete mStartPos "+mStartPos + "mStartSamples" +mStartSamples +
            // "mStartPosTem" + mStartPosTem);
            logHelper.debugShow("OnseekComplete mStartPos " + mStartPos + "mStartPosTem" + mStartPosTem);
        } else {
            mStartPosTem = 0;
            logHelper.warnShow("OnseekComplete error ");
        }
    }

    public void OnPlayComplete() {
        // if (mRenderhandler != null) {
        // mRenderhandler.removeMessages(RENDER_STOP);
        // mRenderhandler.sendEmptyMessage(RENDER_STOP);
        // }
        // AudioTrack audioTrack = mAudioTrack;
        // if (audioTrack != null) {
        // audioTrack.pause();
        // }
        // mIsPlaying = false;
    }

}