package com.baidu.musicRecorder;

import android.media.AudioFormat;
//import android.media.audiofx.*;   // added by ghy@test
import android.media.AudioManager;
import android.media.AudioRecord;
import android.media.MediaRecorder;

import com.baidu.utility.LogHelper;
import com.utility.publicInterface.RecordListener;

public class RecorderManager implements Runnable {
    private final static String tag = "RecorderManager";
    private LogHelper logHelper = new LogHelper(tag);

    private final static int nSleepTime = 10;
    private static final int nAudioSource = MediaRecorder.AudioSource.DEFAULT;
    private final static RecorderManager instance = new RecorderManager();

    private int bufferSizeInBytes = 0;

    private int nSampleRate = 44100;
    private int nChannelNum = AudioFormat.CHANNEL_IN_MONO;
    private int nAudioFormat = AudioFormat.ENCODING_PCM_16BIT;

    private boolean bIsAlive = true;

    private int nThreadRef = 0;
    private boolean bStartRecord = false;
    private int nRecordSize = 8 * 1024;

    private byte[] buffer = new byte[nRecordSize];

    private RecordListener mRecordListener = null;
    private Thread mWorkThread = null;
    private AudioRecord mAudioRecord = null;
    // private AudioTrack mAudioTrack = null;

    private boolean mSingWhenRecording = false;

    private int mStreamtype = AudioManager.STREAM_MUSIC;
    private int mTotalRecordByte = 0;

    private RecorderManager() {
    }

    public final static RecorderManager sharedPreferenceRecorderManager() {
        return instance;
    }

    public void setSingWhenRecording(boolean mSingWhenRecording) {
        logHelper.infoShow("mSingWhenRecording " + mSingWhenRecording);
        this.mSingWhenRecording = mSingWhenRecording;
    }

    public void setRecordListener(RecordListener mListener) {
        this.mRecordListener = mListener;
    }

	public void setRecordBufferSize(int size) {
	    logHelper.infoShow("setRecordBufferSize " + size);
	    if (size != nRecordSize) {
	        nRecordSize = size;
	        buffer = new byte[nRecordSize];
        }
	}
    public void setRecordSettings(int nSampleRate, int nChannelNum, int nAuduiFormat) {
        this.nSampleRate = nSampleRate;
        this.nChannelNum = nChannelNum;
        this.nAudioFormat = nAuduiFormat;
    }

    private void createRecordResource() {
        releaseRecordResource();
        bufferSizeInBytes = AudioRecord.getMinBufferSize(nSampleRate, nChannelNum, nAudioFormat);
        logHelper.debugShow("AudioRecord.getMinBufferSize:" + bufferSizeInBytes);
        bufferSizeInBytes *= 4;
        if (bufferSizeInBytes == AudioRecord.ERROR_BAD_VALUE) {
            if (mRecordListener != null)
                mRecordListener.onRecordErrorNotify(RecordError.AUDIO_RECORD_PARAM_INVALID);
        } else if (bufferSizeInBytes == AudioRecord.ERROR) {
            if (mRecordListener != null) {
                mRecordListener.onRecordErrorNotify(RecordError.AUDIO_RECORD_PARAM_UNSUPPORTED);
            }
        } else {
            if (bufferSizeInBytes < 4 * nRecordSize) {
                bufferSizeInBytes = 4 * nRecordSize;
            }

            logHelper.debugShow("AudioRecord set MinBufferSize:" + bufferSizeInBytes);

            mAudioRecord = new AudioRecord(nAudioSource, nSampleRate, nChannelNum, nAudioFormat, bufferSizeInBytes);
            // int channelConfig = AudioFormat.CHANNEL_OUT_MONO;
            // int minPlayBufSize = AudioTrack.getMinBufferSize(nSampleRate,
            // channelConfig, nAudioFormat);
            //
            // if(minPlayBufSize < bufferSizeInBytes) {
            // minPlayBufSize = bufferSizeInBytes;
            // }
            //
            // mAudioTrack = new AudioTrack(mStreamtype, nSampleRate,
            // channelConfig, nAudioFormat, minPlayBufSize,
            // AudioTrack.MODE_STREAM);
            // if (mAudioTrack != null) {
            // mAudioTrack.play();
            // }
        }
    }

    private void releaseRecordResource() {
        if (mAudioRecord != null) {
            if (mAudioRecord.getState() == AudioRecord.STATE_INITIALIZED)
                mAudioRecord.stop();
            mAudioRecord.release();
            mAudioRecord = null;
            logHelper.infoShow("releaseRecordResource");
        }
        // if (mAudioTrack != null) {
        // mAudioTrack.stop();
        // mAudioTrack.release();
        // mAudioTrack = null;
        // }
        // mSingWhenRecording = false;
    }

    private void createBaseResource() {
        killWorkThread();
        createWorkThread();
        ++nThreadRef;
    }

    private void createWorkThread() {
        bIsAlive = true;
        logHelper.infoShow("record createWorkThread nThreadRef " + nThreadRef);

        if (nThreadRef == 0) {
            mWorkThread = new Thread(this);
            mWorkThread.start();
        }
    }

    private void killWorkThread() {
        if (nThreadRef == 0 && mWorkThread != null) {
            if (mWorkThread.isAlive()) {
                bIsAlive = false;
                try {
                    mWorkThread.join();
                } catch (InterruptedException e) {
                    // TODO 自动生成的 catch 块
                    e.printStackTrace();
                }
                mWorkThread = null;
            }
        }
    }

    public void create() {
        createBaseResource();
        mTotalRecordByte = 0;
        logHelper.infoShow("RecordManager create");
    }

    public void destroy() {
        bStartRecord = false;
        --nThreadRef;
        synchronized (this) {
            releaseRecordResource();
        }
        killWorkThread();
        logHelper.infoShow("RecordManager destroy");
    }

    public boolean prepareRecord() {
        boolean rst = false;
        try {
            createRecordResource();
            if (mAudioRecord != null) { // mAudioRecord为空时createRecordResource已上报错误
                if (mAudioRecord.getState() == AudioRecord.STATE_INITIALIZED) {
                    rst = true;
                } else {
                    if (mRecordListener != null) {
                        mRecordListener.onRecordErrorNotify(RecordError.AUDIO_RECORD_UNINIT);
                    }
                }
            }
        } catch (IllegalStateException exception) {
            if (mRecordListener != null) {
                mRecordListener.onRecordErrorNotify(RecordError.AUDIO_RECORD_PARAM_UNSUPPORTED);
            }
            logHelper.errorShow(exception.getMessage());
        }
        return rst;
    }

    public void startRecord() {
        logHelper.debugShow("startRecord begin");
        if (mAudioRecord != null) {
            mAudioRecord.startRecording();
	        bStartRecord = true;
        }
        logHelper.debugShow("startRecord end");
    }

    public void pauseRecord() {
        logHelper.debugShow("pauseRecord begin");
        bStartRecord = false;
        if (mAudioRecord != null) {
            mAudioRecord.stop();
        }
        logHelper.debugShow("pauseRecord end");

    }

    public void stopRecord() {
        logHelper.infoShow("stopRecord");
        bStartRecord = false;
        synchronized (this) {
            logHelper.infoShow("stopRecord enter");

            releaseRecordResource();
        }
        mTotalRecordByte = 0;
    }

    public int getTotalRecordMs() {
        int nChannel = (nChannelNum == AudioFormat.CHANNEL_IN_STEREO ? 2 : 1);
        int nByteDepth = (nAudioFormat == AudioFormat.ENCODING_PCM_16BIT ? 2 : 1);

        return (int) (((long) mTotalRecordByte) * 1000 / (nSampleRate * nChannel * nByteDepth));
    }

    private void sleepTime(long nMills) {
        try {
            Thread.sleep(nMills);
        } catch (InterruptedException e) {
            // TODO 自动生成的 catch 块
            e.printStackTrace();
        }
    }

    private static final int MAX_NO_DATA_COUNT = 320; // 320 * 50ms

    @Override
    public void run() {
        // android.os.Process.THREAD_PRIORITY_URGENT_AUDIO is too high,
        // some synchronized may be blocked
        // int nTid = android.os.Process.myTid();
        // int nOriPriorty = android.os.Process.getThreadPriority(nTid);
        // // 设置线程优先级，保证资源分配。
        // android.os.Process
        // .setThreadPriority(android.os.Process.THREAD_PRIORITY_URGENT_AUDIO);
        //
        // int nCurPriorty = android.os.Process.getThreadPriority(nTid);
        //
        // logHelper.infoShow("RenderHandlerThreadPriority nTid "+ nTid +
        // "nOriPriorty " + nOriPriorty + "nCurPriorty" + nCurPriorty);

        int nNoDataCount = 0;
        // TODO 自动生成的方法存根
        int nReadSize = 0;
        while (bIsAlive) {
            if (mAudioRecord != null && bStartRecord) {
                synchronized (RecorderManager.this) {
                    if (mAudioRecord != null && bStartRecord) {
                        if ((nReadSize = mAudioRecord.read(buffer, 0, nRecordSize)) 
                                == AudioRecord.ERROR_INVALID_OPERATION) {
                            nReadSize = 0;
                            logHelper.infoShow("mAudioRecord error");
                            if (mRecordListener != null)
                                mRecordListener.onRecordErrorNotify(RecordError.AUDIO_RECORD_INVALID_OPERATION);
                            continue;
                        } else if (0 == nReadSize) {
                            ++nNoDataCount;
                            if (nNoDataCount > MAX_NO_DATA_COUNT) {
                                nNoDataCount = 0;
                                logHelper.infoShow("mAudioRecord error no data");
                                if (mRecordListener != null) {
                                    mRecordListener.onRecordErrorNotify(RecordError.AUDIO_RECORD_NODATA);
                                }
                            }
                            sleepTime(nSleepTime * 5);
                            logHelper.infoShow("mAudioRecord no data " + nNoDataCount);
                        } else {
                            nNoDataCount = 0;
                        }
                    }
                }

                mTotalRecordByte += nReadSize;
                // logHelper.debugShow(" mAudioRecord.read:" + nReadSize +
                // "mTotalRecordByte:" + mTotalRecordByte);
                // if (mSingWhenRecording && mAudioTrack != null) {
                // mAudioTrack.write(buffer, 0, nReadSize);
                // logHelper.infoShow("buffer " + nReadSize);
                // }

                if (true) {
                    // if (mTotalRecordByte > bufferSizeInBytes) {
                    if (mRecordListener != null) {
                        // logHelper.debugShow("in onRecordDataDidRead");
                        mRecordListener.onRecordDataDidRead(buffer, nReadSize);
                    }
                }
                // logHelper.debugShow(" onRecordDataDidRead");
            } else {
                sleepTime(nSleepTime);
            }
        }
    }
}
