package com.baidu.util.audiocore;

import java.io.File;
import java.io.FileDescriptor;
import java.io.IOException;
import java.lang.ref.WeakReference;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import android.content.BroadcastReceiver;
import android.content.ContentResolver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.content.res.AssetFileDescriptor;
import android.media.AudioFormat;
import android.media.AudioManager;
import android.media.AudioTrack;
import android.net.ConnectivityManager;
import android.net.NetworkInfo;
import android.net.Uri;
import android.os.Build;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.os.Handler.Callback;
import android.os.HandlerThread;
import android.os.Looper;
import android.os.Message;
import android.os.PowerManager;
import android.os.StatFs;
import android.os.SystemClock;
import android.telephony.TelephonyManager;

import com.baidu.util.advertisement.AdvertiseMent;
import com.baidu.util.advertisement.AdvertiseMentListInfo;
import com.baidu.util.advertisement.AdvertiseMentMulti;
import com.baidu.util.advertisement.AdvertiseMentMultiListInfo;
import com.baidu.util.advertisement.AdvertiseMentPolicy;
import com.baidu.util.advertisement.Anchor;
import com.baidu.util.advertisement.Noise;
import com.baidu.utility.LogHelper;

/**
 * @author lg
 *
 */
public class AudioPlayer {

    private static Map<String, OnTaskDownloadDelegate> taskDownloadDelegates =
            new HashMap<String, OnTaskDownloadDelegate>();

    private static Object lock = new Object();
    private static Object lockAd = new Object();
    // // public static final boolean LOGGABLE = true; // debug version
    //
    // public static final boolean LOGGABLE = false; //release version

    public static final boolean WMAGABLE = true; // enable wma
    // public static final boolean WMAGABLE = false; //disable wma

    // public static final boolean INSANEGABLE = true; // enable insane
    public static final boolean INSANEGABLE = false; // disable insane

    private static final int TRUE = 1;
    private static final int FALSE = 0;

    private static final boolean LEARN_MODE = true;
    static {
        System.loadLibrary("coreffmpeg");
        if (LEARN_MODE) {
        	System.loadLibrary("ijksoundtouch");
            System.loadLibrary("basecore");
        }
        System.loadLibrary("audiocore");
        native_init();
    }

    private final static String TAG = "AudioPlayer";
    public static String log_PathString = ""; // logcat log if empty
    private int mNativeContext; // accessed by native methods
    private int mListenerContext; // accessed by native methods
    private EventHandler mEventHandler;
    private PowerManager.WakeLock mWakeLock = null;
    private boolean mStayAwake;
    private static final LogHelper logHelper = new LogHelper(TAG);

    // for render
    private static final int RENDER_START = 1;
    private static final int RENDER_RUN = 2;
    private static final int RENDER_STOP = 3;

    private static final int PRI_DOWNLOAD = 2;
    private static final int PRI_PREDOWNLOAD = 3;

    // for network type
    private static final int NETWORK_TYPE_NONE = 0;
    private static final int NETWORK_TYPE_UNKNOWN = 1;
    private static final int NETWORK_TYPE_WIFI = 2;
    private static final int NETWORK_TYPE_2G = 3;
    private static final int NETWORK_TYPE_3G = 4;
    private static final int NETWORK_TYPE_4G = 5;
    // private static final int NETWORK_TYPE_WLAN = 6;//目前没有使用

    private static int mNetState = NETWORK_TYPE_UNKNOWN;
    private static boolean mEnableMobileDataP2P = true;

    private static final int NETWORK_NONE = 0; // 不进行下载上传
    private static final int NETWORK_HTTP_DOWNLOAD = 0x0001; // http下载
    private static final int NETWORK_P2P_DOWNLOAD = 0x0002; // p2p下载
    private static final int NETWORK_P2P_UPLOAD = 0x0004; // p2p上传

    private static Context longLifeContext = null;
    private static boolean isRegisterBroadcast = false;

    int mStreamtype = AudioManager.STREAM_MUSIC;

    private AudioTrack mAudioTrack;
    private RenderHandlerThread mRenderThread;
    private Handler mRenderhandler;
    private int mBufferSize = 0;
    private int mStartPos = 0;
    private int mStartPosTem = 0;
    // private int mStartSamples = 0;
    private int mSessionID = 0;
    private int mSampleRateInHz = 44100;
    private long mAudioTrackSum = 0;
    private long mAudioTrackSumStart = 0;
    private long mAudioTrackSumOriginal = 0;
    private int mBytePerSec = 0;
    private boolean mIsPrepared = false;
    private int mDuration = 0; // millisec
    private boolean mIsEnableFadeInFadeOut = true;
    private boolean mIsPrepareAsync = true;

    private boolean mIsPlaying = false;

    private AdvertiseMentInfos mAdInfos = null;// mingliang ;
    private AdvertiseMentInfoReport mAdReport = null;
    private AdvertiseMentPolicyInline mPolicy = null;

    private List<Integer> mTaskIdList = null;

    private long mAudioTrackSumLast = 0;
    private boolean mAdCurAnchorValid = false;// default false; if debug set it
                                              // true;
    private int mAdCurAdIndex = 0;
    private long mAdCurAdDurationSize = 0;
    private long mAdAbsoluteDurationSize = 0;

    private int mAdCurAnchorIndex = 0;
    private long mAdCurAnchorSize = 0;
    /*
     * multis ad;
     */
    private long mAdMultiAdDurationSize[];
    private long mAdMultiAbsoluteDurationSize[];

    /*
     * debug mAdTimerRelativeMaxTime s play ad ;
     */
    private long mAdTimerRelativeMaxTime = 0;// second;
    private boolean mIsInsertImmediately = false; // 立刻插入广告，所有锚点插入广告，成功插入后标记保持
    private boolean mIsInsertImmediatelyOnce = false; // 是否要立刻插入一次广告，该次广告插入成功后标记该回false
    private static final int mAdPreDownloadTime = 30;// app default 30s
                                                     // //提前通知底层开始下载
    private long mAdPreDownloadSize = 0;

    private static final int mAdPreInsertTime = 0; // app default 10s
                                                   // //提前通知底层开始插入
    private long mAdPreInsertSize = 0;

    private static final int mAdPrePreDownloadTime = 60; // app default 60s
                                                         // //提前通知上层开始获取广告物料
    private long mAdPrePreDownloadSize = 0;

    private long mAdTimerRelativeMaxSize = 0;
    private long mAdTimerRelativeSize = 0;
    private long mAdTimerRelativeLastSize = 0;

    // private int Absolute Timer
    private long mAdTimerSize = 0; // byte
    private long mAdTimerSizeTmp = 0; // byte

    private boolean mAdSingalPreDownloadValid = true;
    private boolean mAdSingalDownloadValid = true;
    private boolean mAdSingalInsertValid = true;
    /*
     * if ad playing ,the mAdhavePlayedValid is true;
     */
    private boolean mAdhavePlayedValid = false;
    private boolean mAdWillPlayValid = false;
    private boolean mAdReportMessageValid = true;

    private boolean mAdFirstInsertValid = true;
    // insert condition;
    private boolean mAdInsertConditionSingalSongMaxCountIsValid = false;

    // 额外预留空间
    private static final int EXTRA_SPACE_BYTE = 10 * 1024 * 1024;

    private float mVolumeLeft = 1.0f;
    private float mVolumeRight = 1.0f;

    // 播放速度0.5f - 2.0f
    private float mPlaybackSpeed = 1.0f;
    /**
     * set context, call this method before the instance of Audioplayer is created.
     */
    public static void setLongLifeContext(Context context) {
        longLifeContext = context;
        if (!isRegisterBroadcast && longLifeContext != null) {
            IntentFilter intentFilter = new IntentFilter();
            intentFilter.addAction(ConnectivityManager.CONNECTIVITY_ACTION);
            longLifeContext.registerReceiver(connectionReceiver, intentFilter);
            isRegisterBroadcast = true;
        }
    }

    /**
     * release context.
     */
    public static void releaseLongLifeContextResource() {
        if (longLifeContext != null && isRegisterBroadcast) {
            try {
                longLifeContext.unregisterReceiver(connectionReceiver);
            } catch (IllegalArgumentException e) {
                e.printStackTrace();
            }
            isRegisterBroadcast = false;
            longLifeContext = null;
        }
    }

    /**
     * Sets path to store native crash info.
     * 
     * @param strFilePath the folder to store native crash info, when crashed ****.dmp
     * @throws IllegalArgumentException
     */
    public static void setNativeCrashFilePath(String strFilePath) throws IllegalArgumentException {
        logHelper.debugShow("setNativeCrashFilePath " + strFilePath);
        native_setNativeCrashFilePath(strFilePath);
    }

    /**
     * Constructor.
     * 
     * @param eventlooper out looper, or null 
     */
    public AudioPlayer(Looper eventlooper) {
        if (eventlooper != null) {
            logHelper.infoShow("AudioPlayer eventlooper out");
            mEventHandler = new EventHandler(this, eventlooper);
        } else {
            logHelper.infoShow("AudioPlayer eventlooper null");
            Looper looper;
            if ((looper = Looper.myLooper()) != null) {
                mEventHandler = new EventHandler(this, looper);
            } else if ((looper = Looper.getMainLooper()) != null) {
                mEventHandler = new EventHandler(this, looper);
            } else {
                mEventHandler = null;
            }
        }

        // Create render thread.
        mRenderThread = new RenderHandlerThread("render");
        mRenderThread.start();
        mRenderhandler = new Handler(mRenderThread.getLooper(), mRenderThread);

        /*
         * Native setup requires a weak reference to our object. It's easier to create it here than in C++.
         */
        native_setup(new WeakReference<AudioPlayer>(this));

        CreateDownloadTaskHandler();

        // int flag = ((LOGGABLE) ? 1 : 0);
        // flag += ((WMAGABLE) ? 0 : 2);
        // flag += ((INSANEGABLE) ? 4 : 0);
        // setPlayerFlag(flag);
        // logHelper.EnableLog(LOGGABLE);

    }

    public AudioPlayer() {
        this(null);
    }

    /**
     * enable log
     * @param loggable enable log or disable log 
     */
    public static void enableGlobalLog(boolean loggable) {
        logHelper.enableLog(loggable);
        int flag = ((loggable) ? 1 : 0);
        flag += ((WMAGABLE) ? 0 : 2);
        flag += ((INSANEGABLE) ? 4 : 0);
        setPlayerFlag(flag);
    }

    // p2p begin

    // auth type, used by setProxy 
    public static final int AC_AUTH_TYPE_BASIC = 0;
    public static final int AC_AUTH_TYPE_DIGEST = 1;

    /**
     * set proxy, if not need strUser and strPwd, you can set them nil
     * @param host host of proxy 
     * @param port port of proxy 
     * @param user username of proxy 
     * @param passwd password of proxy 
     */
    public static void setProxy(String host, int port, String user, String passwd) {
    	native_setProxy(host, port, user, passwd, AC_AUTH_TYPE_BASIC);
    }


    /**
     * set proxy, if not need strUser and strPwd, you can set them nil
     * @param host host of proxy 
     * @param port port of proxy 
     * @param user username of proxy 
     * @param passwd password of proxy 
     * @param authType auth type: Basic(AC_AUTH_TYPE_BASIC) or Digest (AC_AUTH_TYPE_DIGEST)
     */
    public static void setProxy(String host, int port, String user, String passwd, int authType) {
    	native_setProxy(host, port, user, passwd, authType);
    }

    /**
     * whether make p2p upload resource,useless
     * @param bUse 设置是否允许上传 
     */
    public static native void isUseUpload(boolean bUse);

    /**
     * download percent, total percent is 100
     * @param IsProgress false needed 
     * @return download progress (0~100)
     */
    public native int getProgress(boolean IsProgress);

    public static final int PID_MAIN_MUSIC = 0x04000000;
    public static final int PID_MAIN_LEBO = 0x08000000;
    public static final int PID_SDK_MUSIC = 0x0A000000;
    public static final int PID_SDK_LEBO = 0x0B000000;
    public static final int PID_SHOUBAI_MUSIC = 0x0D000000;
    public static final int PID_YYR = 0x0F000000;
    public static final int PID_LAVA = 0x11000000;

    protected static int s_pid = 0;
    
    /**
     * global initial net environment, call this method only once, call this method before 
     * AudioPlayer is created, call this method before download or predownload
     * @param cacheFolder net cache folder 
     * @param uCacheSize cache size (MB)
     * @param pId etc PID_SDK_MUSIC 0x0AXXXXXX,第一个字节传入0x0A 用于标记android SDK，后面24位记录3位版本号
     * ，如1.2.0及 0x08010200
     * @param dId always 0
     * @return success or fail
     */
    public static boolean initP2pEnvironment(String cacheFolder, int uCacheSize /* MB */, int pId, int dId) {
        if (cacheFolder == null || cacheFolder.trim().equals("") || uCacheSize < 0 || pId < 0 || dId < 0) {
            return false;
        }
        AudioPlayer.setCacheFolder(cacheFolder, uCacheSize);
        AudioPlayer.setDeviceID(pId, dId);

        s_pid = pId;

        return true;
    }
    
    /**
     * init enviroment, 内部会调用获取设备信息的方法，所以需要在获取读取设备状态的权限后再调用。
     * 如果没有获取到权限也可以调用该方法，上报的部分信息缺失
     */
    public static void setEnvironmentInfo() {
        setEnvironmentInfo(s_pid);
    }
    
    /**
     * post a download file task, return task id if success
     * @param url url to download
     * @param delegate monitor download process
     * @return task id, if success, return value not equals 0
     */
    public static int postDownloadTask(String url, OnTaskDownloadDelegate delegate) { // post a task of download
        return postTask(PRI_DOWNLOAD, url, delegate);
    }

    /**
     * post a pre download file task, return task id if success.prior is not high as postDownloadTask.
     * @param url url to download
     * @param delegate monitor download process
     * @return task id, if success, return value not equals 0
     */
    public static int postPreDownloadTask(String url, OnTaskDownloadDelegate delegate) { // post a task of pre download
        return postTask(PRI_PREDOWNLOAD, url, delegate);
    }
    
    // 开启预下载,同时告知正在进行下载的任务Id，如果成功,返回一个任务Id
    public static int postPreDownloadTask(int fatherTaskId, String url, OnTaskDownloadDelegate delegate) {
        int sonTaskId = 0;
        sonTaskId = postTask(PRI_PREDOWNLOAD, url, delegate);
        if (fatherTaskId != 0) {
            relationTasks(fatherTaskId, sonTaskId);
        }
        return sonTaskId;
    }

    /**
     * clean net cache, you can alse delete cache folder set by initP2pEnvironment manually.
     * @return spacce freed by this method, in bytes
     */
    public static native int clearNetCache();

    // 获得当前缓存大小，单位byte,仅供参考，当前缓存刷新可能不及时
    /**
     * get curent cache size, in bytes. only for reference as it is not realtime.
     * @return cache size, in bytes
     */
    public static native int getCurrentCacheFolderSize();

    // get file path of download or predownload
    
    /**
     * get file path of download or pre download.
     * @param nTaskId task id from postPreDownloadTask or postPreDownloadTask
     * @return file path in cache folder
     */
    public static native String getDownloadFilePathString(int nTaskId);

    // delete a task
    /**
     * delete a task.
     * @param nTaskId task id from postPreDownloadTask or postPreDownloadTask
     * @return success or fail
     */
    public static boolean deleteDownloadTask(int nTaskId) {
        if (nTaskId > 0) {
            if (deleteTask(nTaskId) == TRUE) {
                String strTaskIdString = String.valueOf(nTaskId);
                if (taskDownloadDelegates.containsKey(strTaskIdString)) {
                    taskDownloadDelegates.put(strTaskIdString, null);
                    taskDownloadDelegates.remove(strTaskIdString);
                }
                return true;
            }
        }
        return false;
    }

    
    /**
     * stop current download task
     * @throws IllegalStateException
     */
    public void stopCurrentTask() throws IllegalStateException {
        native_stopCurrentTask();
    }

    
    /**
     * call initP2pEnvironment will be ok
     * @param path net cache folder 
     * @param size cache size (MB)
     */
    public static native void setCacheFolder(String path, int size);

    // call initP2pEnvironment will be ok
    
    /**
     * call initP2pEnvironment will be ok
     * @param ProductID etc PID_SDK_MUSIC 0x0AXXXXXX,第一个字节传入0x0A 用于标记android SDK，后面24位记录3位版本号
     * ，如1.2.0及 0x08010200
     * @param DeviceID always 0
     */
    public static native void setDeviceID(int ProductID, int DeviceID);
    
    /**
     * set user agent, should be called after initP2pEnvironment is called
     * @param ua user agent
     * @throws IllegalArgumentException
     */
    public static void setUserAgent(String ua) throws IllegalArgumentException {
        logHelper.debugShow("setUserAgent:" + ua);
        native_setUserAgent(ua);
    }

    /**
     * set uniqueIdentifier of partner, music SDK or lebo SDK should call this method, 
     * should be called after initP2pEnvironment is called
     * @param partner uniqueIdentifier
     * @throws IllegalArgumentException
     */
    public static void setPartner(String partner) throws IllegalArgumentException {
        logHelper.debugShow("setPartner:" + partner);
        native_setPartner(partner);
    }

    
    /**
     * set user custom http header, now used by China Mobile 
     * @param keys etc "ClientID" or "ClientID@@@@ClientVer"
     * @param values etc "MyClient31-User67328" or"MyClient31-User67328@@@@1.0"
     * @param seperator etc "" or "@@@@"
     * @throws IllegalArgumentException
     */
    public static void setCustomHttpHeader(String keys, String values, String seperator)
            throws IllegalArgumentException {
        logHelper.debugShow("setCustomHttpHeader keys:" + keys 
                + " values: " + values + " seperator: " + seperator);
        native_setCustomHttpHeader(keys, values, seperator);
    }
    
    /**
     * net path should be called after prepare complete
     * @return 
     * @throws IllegalStateException
     */
    public String getCurFilePathString() throws IllegalStateException {
        return native_getCurFilePathString();
    }

    /**
     * set whether use p2p when current net state is mobile(3G/4G). current is used by flow free.
     * 
     * @param bEnable true enable p2p, false disable p2p
     */
    public static void enableMobileDataP2P(boolean bEnable) {
        logHelper.debugShow("enableMobileDataP2P:" + bEnable);
        mEnableMobileDataP2P = bEnable;
        setUpDownloadTypeByNetType();
    }

    /**
     * set download task timeout time. will receive OnTaskDownloadDelegate::onTaskIdTimeout when download timeout
     * 
     * @param timeoutseconds
     */
    public static void setDownloadTaskTimeout(int timeoutseconds) {
        logHelper.debugShow("setDownloadTaskTimeout:" + timeoutseconds);
        native_setDownloadTaskTimeout(timeoutseconds);
    }

    /**
     * url是否已经缓存
     * @param httpurl 原生可以下载的http url 如http://xxx.mp3，不是组装好的p2p://XXXX
     * @param hash 该链接对应文件的hash值，没有可以传入 null
     * @return 是否已经缓存
     */
    public static boolean isUrlCached(String httpurl, String hash) {
        boolean iscached = native_isUrlCached(httpurl, hash);
        logHelper.debugShow("isUrlCached httpurl:" + httpurl + " iscached: " + iscached);
        return iscached;
    }
    
    // 不用设置
    public static native void setSpeedMode(int mode);

    // 不用设置
    public static native void setTaskPriority(int uPriority, int uTaskID);

    // 不用设置
    public static native void setMaxDownloadSpeed(int uSpeed, int uTaskID);

    // 不用设置
    public static native void setNetType(int type);

    // p2p end

    /**
     * app set important info to audiocore.
     * 
     * @param info setAppInfo
     * @throws IllegalStateException if it is called in an invalid state
     * @throws IOException
     * @throws IllegalArgumentException
     */
    private void _setAppInfo(HashMap<String, String> infoMap) throws IllegalStateException, IllegalArgumentException {
        if (null == infoMap) {
            return;
        }

        String strKey[] = new String[infoMap.size()];
        String strVal[] = new String[infoMap.size()];

        int i = 0;
        Iterator<String> iter = infoMap.keySet().iterator();

        while (iter.hasNext()) {
            strKey[i] = (String) iter.next();
            // logHelper.infoShow(" check--key- : " + i + "key=" + strKey[i]);
            i++;
        }

        i = 0;
        iter = infoMap.values().iterator();
        while (iter.hasNext()) {
            strVal[i] = (String) iter.next();
            // logHelper.infoShow(" check--val- : " + i + "val=" + strVal[i]);
            i++;
        }
        native_setAppInfo(strKey, strVal);
    }

    private native void native_setAppInfo(String key[], String val[]) throws IllegalArgumentException,
            IllegalStateException;

    /**
     * Sets the data source as a content Uri.
     * 
     * @param context the Context to use when resolving the Uri
     * @param uri the Content URI of the data you want to play
     * @throws IllegalStateException if it is called in an invalid state
     */
    public void setDataSource(Context context, Uri uri) throws IOException, IllegalArgumentException,
            SecurityException, IllegalStateException {
        logHelper.infoShow("setDataSource start ");
        String scheme = uri.getScheme();
        if (scheme == null || scheme.equals("file")) {
            setDataSource(uri.getPath());
            return;
        }

        AssetFileDescriptor fd = null;
        try {
            ContentResolver resolver = context.getContentResolver();
            fd = resolver.openAssetFileDescriptor(uri, "r");
            if (fd == null) {
                logHelper.infoShow("setDataSource end fd=null");
                return;
            }
            // Note: using getDeclaredLength so that our behavior is the same
            // as previous versions when the content provider is returning
            // a full file.
            if (fd.getDeclaredLength() < 0) {
                setDataSource(fd.getFileDescriptor());
            } else {
                setDataSource(fd.getFileDescriptor(), fd.getStartOffset(), fd.getDeclaredLength());
            }
            return;
        } catch (SecurityException ex) {
        } catch (IOException ex) {
        } finally {
            if (fd != null) {
                fd.close();
            }
        }
        logHelper.debugShow("Couldn't open file on client side, trying server side");
        setDataSource(uri.toString());
        logHelper.infoShow("setDataSource end ");
        return;
    }

    /**
     * Sets the data source (file-path or http/rtsp URL) to use.
     * 
     * @param path the path of the file, or the http/rtsp URL of the stream you want to play infomap the info for track
     *            problem (key-val)
     * @throws IllegalStateException if it is called in an invalid state
     */
    public void setDataSource(String path, HashMap<String, String> infomap) throws IOException,
            IllegalArgumentException, IllegalStateException {
        setDataSource(path);
        _setAppInfo(infomap);
    }

    /**
     * Sets the data source (file-path or http/rtsp URL) to use.
     * 
     * @param path the path of the file, or the http/rtsp URL of the stream you want to play
     * @throws IllegalStateException if it is called in an invalid state
     */
    public void setDataSource(String path) throws IOException, IllegalArgumentException, IllegalStateException {

        // 切歌时上层并没有调用reset导致广告播放状态可能没有重置
        mAdReport = new AdvertiseMentInfoReport();

        if (mAdReport != null) {
            mAdReport.singal = AD_UNKNOWN;
            mAdReport.curIndex = 0;
            mAdReport.timerAbsoluteSumMs = 0;
        }

        native_setDataSource(path);
        if (mAdWillPlayValid) {

            // report audiocore;
            formatAdSendSignal(AD_WILLPLAY);

            logHelper.debugShow("setDataSource normal mAdWillPlayValid=true");
            resetAdStatus();
        }
    }

    /**
     * Sets the data source (FileDescriptor) to use. It is the caller's responsibility to close the file descriptor. It
     * is safe to do so as soon as this call returns.
     * 
     * @param fd the FileDescriptor for the file you want to play
     * @throws IllegalStateException if it is called in an invalid state
     */
    public void setDataSource(FileDescriptor fd) throws IOException, IllegalArgumentException, IllegalStateException {
        // intentionally less than LONG_MAX
        setDataSource(fd, 0, 0x7fffffffL);
    }

    /**
     * Sets the data source (FileDescriptor) to use. It is the caller's responsibility to close the file descriptor. It
     * is safe to do so as soon as this call returns.
     * 
     * @param fd the FileDescriptor for the file you want to play
     * @param offset the offset into the file where the data to be played starts, in bytes
     * @param length the length in bytes of the data to be played
     * @throws IllegalStateException if it is called in an invalid state
     */
    public native void setDataSource(FileDescriptor fd, long offset, long length) throws IOException,
            IllegalArgumentException, IllegalStateException;

    /**
     * set the advertisementinfo to player
     * 
     * @param info AdvertiseMentListInfo
     *        infomap (key-val string-string)
     * @throws IllegalStateException if it is called in an invalid state
     */
    public void setDataSource(AdvertiseMentMultiListInfo info, HashMap<String, String> infomap)
            throws IllegalStateException, IllegalArgumentException, IOException {
        setDataSource(info);
        _setAppInfo(infomap);
    }

    /**
     * set the advertisementinfo to player
     * 
     * @param info AdvertiseMentListInfo
     * @throws IllegalStateException if it is called in an invalid state
     */
    public void setDataSource(AdvertiseMentMultiListInfo info) throws IllegalStateException {

        if ((info != null) && (info.mediaUrl != null)) {
            mAdInfos = new AdvertiseMentInfos();
            mAdInfos.mediaUrl = info.mediaUrl;
            mAdInfos.mediaDuration = info.mediaDuration;
        } else {
            mAdInfos = null;
            return;
        }

        if (mAdInfos != null) {
            // get total nums of ad;
            int NumsOfAd = 0;

            // check to find non-null adlist
            if (info.adListAdvertiseMentMulti != null) {
                List<AdvertiseMent> adListExist = null;
                for (int i = 0; i < info.adListAdvertiseMentMulti.size(); i++) {
                    AdvertiseMentMulti mMulti = info.adListAdvertiseMentMulti.get(i);
                    if (mMulti.adListAdvertiseMent != null) {
                        adListExist = mMulti.adListAdvertiseMent;
                        break;
                    }
                }

                for (int w = 0; w < info.adListAdvertiseMentMulti.size(); w++) {
                    AdvertiseMentMulti mMulti = info.adListAdvertiseMentMulti.get(w);
                    if (mMulti.adListAdvertiseMent != null) {
                        NumsOfAd += mMulti.adListAdvertiseMent.size();
                    }

                    // if cur adlist is null, set non-null adlist to cur anchor
                    // pos
                    if ((adListExist != null)
                            && ((null == mMulti.adListAdvertiseMent) || ((mMulti.adListAdvertiseMent != null) && (0 == mMulti.adListAdvertiseMent
                                    .size())))) {
                        mMulti.adListAdvertiseMent = adListExist;
                        info.adListAdvertiseMentMulti.set(w, mMulti);
                        NumsOfAd += mMulti.adListAdvertiseMent.size();
                    }
                }
            }

            // 没有广告要插入则所有播放开始时插入策略（冠名广告）失效
            if ((0 == NumsOfAd) && mIsInsertImmediatelyOnce) {
                mIsInsertImmediatelyOnce = false;
                logHelper.warnShow("setDataSource no ad but need insert at fisrt play");
            }

            // multi ads / anchor
            // if (NumsOfAd > mAdInfos.anchorNums){
            if (NumsOfAd > 0) {
                mAdMultiAdDurationSize = new long[NumsOfAd];
                mAdMultiAbsoluteDurationSize = new long[NumsOfAd];
            }

            // find anchor
            if ((info != null) && (info.adListAdvertiseMentMulti != null) && (info.adListAdvertiseMentMulti.size() > 0)
                    && (mAdInfos != null)) {
                mAdInfos.anchorNums = info.adListAdvertiseMentMulti.size();
                mAdInfos.anchorPos = new int[mAdInfos.anchorNums];
                mAdInfos.anchorTime = new int[mAdInfos.anchorNums];
                mAdInfos.anchorNumsOfAd = new int[mAdInfos.anchorNums];

                mAdInfos.adNums = NumsOfAd;
                mAdInfos.adId = new int[mAdInfos.adNums];
                mAdInfos.adurl = new String[mAdInfos.adNums];
                mAdInfos.adDuration = new int[mAdInfos.adNums];

                int adcount = 0;
                for (int i = 0; i < info.adListAdvertiseMentMulti.size(); i++) {
                    AdvertiseMentMulti mMulti = info.adListAdvertiseMentMulti.get(i);
                    mAdInfos.anchorPos[i] = mMulti.adAnchor.anchorPos;
                    mAdInfos.anchorTime[i] = mMulti.adAnchor.anchorTime;

                    if (mMulti.adListAdvertiseMent != null && (mMulti.adListAdvertiseMent.size() > 0)) {
                        mAdInfos.anchorNumsOfAd[i] = mMulti.adListAdvertiseMent.size();
                        for (int j = 0; j < mMulti.adListAdvertiseMent.size(); j++) {
                            AdvertiseMent mAdMent = new AdvertiseMent();
                            mAdMent = mMulti.adListAdvertiseMent.get(j);
                            mAdInfos.adurl[adcount] = mAdMent.adUrl;
                            mAdInfos.adDuration[adcount] = mAdMent.adDuration;
                            mAdInfos.adId[adcount] = mAdMent.adId;
                            adcount++;
                        }
                    }
                }
            } else { // not found
                mAdInfos.anchorNums = 0;
                mAdInfos.adNums = 0;
            }

            logHelper.infoShow("policy NumsOfAd NumsOfAnchor (" + NumsOfAd + "," + mAdInfos.anchorNums + ")");

            if ((info != null) && (info.adListNoise != null) && (info.adListNoise.size() > 0)) {
                mAdInfos.noiseNums = info.adListNoise.size();
                mAdInfos.noiseStartTime = new int[mAdInfos.noiseNums];
                mAdInfos.noiseEndTime = new int[mAdInfos.noiseNums];
                mAdInfos.noiseStartPos = new int[mAdInfos.noiseNums];
                mAdInfos.noiseEndPos = new int[mAdInfos.noiseNums];

                for (int i = 0; i < info.adListNoise.size(); i++) {
                    Noise mNoise = new Noise();
                    mNoise = info.adListNoise.get(i);
                    mAdInfos.noiseStartTime[i] = mNoise.noiseStarttime;
                    mAdInfos.noiseEndTime[i] = mNoise.noiseEndtime;
                    mAdInfos.noiseStartPos[i] = mNoise.noiseStartpos;
                    mAdInfos.noiseEndPos[i] = mNoise.noiseEndpos;
                }
            } else {
                mAdInfos.noiseNums = 0;
            }

            mAdInfos.normalAnchor();
        }

        mAdCurAnchorValid = false;// default false; if debug set it true;
        mAdReportMessageValid = true;// default true;

        mAdTimerSize = 0;
        mAdTimerSizeTmp = 0;

        mAdCurAdIndex = 0;
        mAdCurAdDurationSize = 0;
        mAdAbsoluteDurationSize = 0;

        mAdCurAnchorIndex = 0;
        mAdCurAnchorSize = 0;

        if (mAdInfos != null && (mAdInfos.mediaUrl != null)) {
            advertisementSetInfos(mAdInfos);
        }

        mAdReport = new AdvertiseMentInfoReport();

        if (mAdReport != null) {
            mAdReport.singal = AD_UNKNOWN;
            mAdReport.curIndex = 0;
            mAdReport.timerAbsoluteSumMs = 0;
        }

        // mIsInsertImmediately first anchor play ad
        if ((mAdWillPlayValid || mIsInsertImmediately || mIsInsertImmediatelyOnce) && (mAdInfos != null)
                && (mAdReport != null) && (mAdInfos.anchorNums > 0)) {
            logHelper.infoShow("policy newmedia adplaying SingalInsert (" + mAdCurAdIndex + "," + mAdCurAnchorIndex
                    + "," + mAdWillPlayValid + "," + mIsInsertImmediatelyOnce + ")");

            // report audiocore;
            formatAdSendSignal(AD_INSERT);

            mAdSingalPreDownloadValid = false;
            mAdSingalDownloadValid = false;
            mAdSingalInsertValid = false;
        }

        if (mAdWillPlayValid || mIsInsertImmediately || mIsInsertImmediatelyOnce) {
            // report audiocore;
            formatAdSendSignal(AD_WILLPLAY);

            if (mOnAdvertiseMentPlayStatusListener != null) {
                mOnAdvertiseMentPlayStatusListener.onAdvertiseMentPlayStatus(AudioPlayer.this, AD_WILLPLAY,
                        ((mAdInfos.adId != null) && (mAdInfos.adNums > 0) ? mAdInfos.adId[0] : -1), 0);
            }
        }

        // 重置立刻插入一次广告
        if (mIsInsertImmediatelyOnce) {
            mIsInsertImmediatelyOnce = false;
        }

        logHelper.infoShow(" policy-setdatasource- firstvalid Maxvalid oncevalid(" + mAdFirstInsertValid + ","
                + mAdInsertConditionSingalSongMaxCountIsValid + "," + mIsInsertImmediatelyOnce + ")");

    }

    /**
     * set the List of AdvertiseMent to player
     * 
     * @param info List<AdvertiseMent>
     * @throws IllegalStateException if it is called in an invalid state
     */
    public void setADDataSource(List<AdvertiseMent> info) throws IllegalStateException {
        if (null == mAdInfos) {
            return;
        }

        if (null == info) {
            return;
        }

        if (0 == info.size()) {
            return;
        }

        logHelper.infoShow("policy setADDataSource ( adDuration " + info.get(0).adDuration + ", adId "
                + info.get(0).adId + ")");

        if (null == mTaskIdList) {
            mTaskIdList = new ArrayList<Integer>();
        }

        for (int i = 0; i < info.size(); ++i) {
            int nTaskId = postPreDownloadTask(info.get(i).adUrl, null);
            if (nTaskId > 0) {
                mTaskIdList.add(nTaskId);
            }
        }

        // get total nums of ad;
        int NumsOfAd = mAdInfos.anchorNums * info.size();

        // set ad
        if (NumsOfAd > 0) {
            mAdMultiAdDurationSize = new long[NumsOfAd];
            mAdMultiAbsoluteDurationSize = new long[NumsOfAd];

            mAdInfos.adNums = NumsOfAd;
            mAdInfos.adId = new int[mAdInfos.adNums];
            mAdInfos.adurl = new String[mAdInfos.adNums];
            mAdInfos.adDuration = new int[mAdInfos.adNums];

            int adcount = 0;
            for (int i = 0; i < mAdInfos.anchorNums; i++) {
                mAdInfos.anchorNumsOfAd[i] = info.size();
                for (int j = 0; j < info.size(); j++) {
                    mAdInfos.adurl[adcount] = info.get(j).adUrl;
                    mAdInfos.adDuration[adcount] = info.get(j).adDuration;
                    mAdInfos.adId[adcount] = info.get(j).adId;
                    adcount++;
                }
            }
        } else { // not found
            mAdInfos.anchorNums = 0;
            mAdInfos.adNums = 0;
        }

        if ((mAdInfos != null) && (mAdInfos.adNums > mAdInfos.anchorNums)) {
            for (int i = 0; i < mAdInfos.adNums; i++) {
                mAdMultiAdDurationSize[i] = (long) ((long) mAdInfos.adDuration[i] * (long) mBytePerSec) / 1000;
                logHelper.infoShow("policy adindex duration(" + i + "," + mAdInfos.adDuration[i] + ","
                        + mAdMultiAdDurationSize[i] + ")");
            }
        }

        if (mAdInfos.mediaUrl != null) {
            advertisementSetAdInfos(mAdInfos);
        }

        logHelper.infoShow(" policy-setdatasource= firstvalid Maxvalid oncevalid(" + mAdFirstInsertValid + ","
                + mAdInsertConditionSingalSongMaxCountIsValid + "," + mIsInsertImmediatelyOnce + ")");
    }

    /**
     * set the advertisementinfo to player
     * 
     * @param info AdvertiseMentListInfo
     * @throws IllegalStateException if it is called in an invalid state
     */
    public void setDataSource(AdvertiseMentListInfo info) throws IllegalStateException {

        if ((info != null) && (info.mediaUrl != null)) {
            mAdInfos = new AdvertiseMentInfos();
            mAdInfos.mediaUrl = info.mediaUrl;
            mAdInfos.mediaDuration = info.mediaDuration;
        } else {
            mAdInfos = null;
            return;
        }

        if (mAdInfos != null) {
            int i = 0;
            if ((info != null) && (info.adListAdvertiseMent != null) && (info.adListAdvertiseMent.size() > 0)
                    && (mAdInfos != null)) {
                mAdInfos.adNums = info.adListAdvertiseMent.size();
                mAdInfos.adId = new int[mAdInfos.adNums];
                mAdInfos.adurl = new String[mAdInfos.adNums];
                mAdInfos.adDuration = new int[mAdInfos.adNums];
                for (i = 0; i < info.adListAdvertiseMent.size(); i++) {
                    AdvertiseMent mAdMent = new AdvertiseMent();
                    mAdMent = info.adListAdvertiseMent.get(i);
                    mAdInfos.adurl[i] = mAdMent.adUrl;
                    mAdInfos.adDuration[i] = mAdMent.adDuration;
                    mAdInfos.adId[i] = mAdMent.adId;
                }
            } else {
                mAdInfos.adNums = 0;
            }

            if ((info != null) && (info.adListAnchor != null) && (info.adListAnchor.size() > 0)) {
                mAdInfos.anchorNums = info.adListAnchor.size();
                mAdInfos.anchorPos = new int[mAdInfos.anchorNums];
                mAdInfos.anchorTime = new int[mAdInfos.anchorNums];
                for (i = 0; i < info.adListAnchor.size(); i++) {
                    Anchor mAnchor = new Anchor();
                    mAnchor = info.adListAnchor.get(i);
                    mAdInfos.anchorPos[i] = mAnchor.anchorPos;
                    mAdInfos.anchorTime[i] = mAnchor.anchorTime;
                }
            } else {
                mAdInfos.anchorNums = 0;
            }

            // 没有广告要插入则所有播放开始时插入策略（冠名广告）失效
            if ((0 == mAdInfos.anchorNums) && mIsInsertImmediatelyOnce) {
                mIsInsertImmediatelyOnce = false;
                logHelper.warnShow("setDataSource no ad but need insert at fisrt play");
            }

            if ((info != null) && (info.adListNoise != null) && (info.adListNoise.size() > 0)) {
                mAdInfos.noiseNums = info.adListNoise.size();
                mAdInfos.noiseStartTime = new int[mAdInfos.noiseNums];
                mAdInfos.noiseEndTime = new int[mAdInfos.noiseNums];
                mAdInfos.noiseStartPos = new int[mAdInfos.noiseNums];
                mAdInfos.noiseEndPos = new int[mAdInfos.noiseNums];

                for (i = 0; i < info.adListNoise.size(); i++) {
                    Noise mNoise = new Noise();
                    mNoise = info.adListNoise.get(i);
                    mAdInfos.noiseStartTime[i] = mNoise.noiseStarttime;
                    mAdInfos.noiseEndTime[i] = mNoise.noiseEndtime;
                    mAdInfos.noiseStartPos[i] = mNoise.noiseStartpos;
                    mAdInfos.noiseEndPos[i] = mNoise.noiseEndpos;
                }
            } else {
                mAdInfos.noiseNums = 0;
            }

            if (mAdInfos != null)
                mAdInfos.normalAnchor();
        }

        mAdCurAnchorValid = false;// default false; if debug set it true;
        mAdReportMessageValid = true;// default true;

        mAdTimerSize = 0;
        mAdTimerSizeTmp = 0;

        mAdCurAdIndex = 0;
        mAdCurAdDurationSize = 0;
        mAdAbsoluteDurationSize = 0;

        mAdCurAnchorIndex = 0;
        mAdCurAnchorSize = 0;

        if (mAdInfos != null && (mAdInfos.mediaUrl != null)) {
            advertisementSetInfos(mAdInfos);
        }

        mAdReport = new AdvertiseMentInfoReport();

        if (mAdReport != null) {
            mAdReport.singal = AD_UNKNOWN;
            mAdReport.curIndex = 0;
            mAdReport.timerAbsoluteSumMs = 0;
        }

        // mIsInsertImmediately first anchor play ad
        if ((mAdWillPlayValid || mIsInsertImmediately || mIsInsertImmediatelyOnce) && (mAdInfos != null)
                && (mAdReport != null) && (mAdInfos.anchorNums > 0)) {
            logHelper.infoShow("policy newmedia adplaying SingalInsert (" + mAdCurAdIndex + "," + mAdCurAnchorIndex
                    + "," + mAdWillPlayValid + "," + mIsInsertImmediatelyOnce + ")");

            formatAdSendSignal(AD_INSERT);

            mAdSingalPreDownloadValid = false;
            mAdSingalDownloadValid = false;
            mAdSingalInsertValid = false;
        }

        if (mAdWillPlayValid || mIsInsertImmediately || mIsInsertImmediatelyOnce) {
            // report audiocore;
            formatAdSendSignal(AD_WILLPLAY);

            if (mOnAdvertiseMentPlayStatusListener != null) {
                mOnAdvertiseMentPlayStatusListener.onAdvertiseMentPlayStatus(AudioPlayer.this, AD_WILLPLAY,
                        ((mAdInfos.adId != null) && (mAdInfos.adNums > 0) ? mAdInfos.adId[0] : -1), 0);
            }
        }

        // 重置立刻插入一次广告
        if (mIsInsertImmediatelyOnce) {
            mIsInsertImmediatelyOnce = false;
        }
        logHelper.infoShow(" policy-setdatasource+ firstvalid Maxvalid oncevalid(" + mAdFirstInsertValid + ","
                + mAdInsertConditionSingalSongMaxCountIsValid + "," + mIsInsertImmediatelyOnce + ")");

    }

    // /*
    // * insert ad unitl play timesecond second, call this method before prepare
    // * timesecond : start to play the advertisement ,relative time ( s )
    // * 0 means insert ads at all anchors
    // */
    // public void advertisementSetStartToPlayRelativeTime(long timesecond) {
    // mAdTimerRelativeMaxTime = timesecond;
    // if (0 == mAdTimerRelativeMaxTime) {
    // mIsInsertImmediately = true;
    // mAdTimerRelativeMaxTime = 60;
    // } else {
    // mIsInsertImmediately = false;
    // }
    // mAdInsertConditionSingalSongMaxCountIsValid = true;
    // }

    /**
     * 在接下来的播放立刻播放一次广告，播放前前调用有效，目前用于冠名广告
     */
    public void playAdOnceImmediately() {
        // 仅在第一次开始播放前可能为true
        mIsInsertImmediatelyOnce = mIsPrepared ? false : true;
        logHelper.infoShow(" policy playAdOnceImmediately mIsInsertImmediatelyOnce:" + mIsInsertImmediatelyOnce);

        // 如果目前没有设置策略，设置默认策略
        if (null == mPolicy) {
            mPolicy = new AdvertiseMentPolicyInline();
            if (mPolicy == null)
                return;

            AdvertiseMentPolicy policy = new AdvertiseMentPolicy();
            policy.adpFirstInsertSongNum = 2;
            policy.adpSingalSongValidTime = 60;
            policy.adpSingalSongMaxCount = 5;
            policy.pMaxTotalTime = 600;
            policy.bNeedCalcAdProgress = true;
            setAdvertisementPolicy(policy);
        }
    }

    /**
     * 设置广告策略.
     */
    public void setAdvertisementPolicy(AdvertiseMentPolicy policy) {
        logHelper.infoShow(" policy-setAdvertisementPolicy");

        // 原来APP在播放广告时切场景调用顺序时setpolicy－reset－setdatasource
        // 导致需要在setpolicy里判断广告中断
        // 现在APP已经修改为reset－setpolicy－setdatasource 所以判断广告中断可以删除.lebo and sdk not
        // support
        // should be called before if(policy == null)
        if (!LEARN_MODE) {
            checkAdInterrupt();
        }

        synchronized (lockAd) {
            // if ad played ,not run (2,3,0) - 2;
            if (mAdhavePlayedValid == false) {
                mAdFirstInsertValid = true;
            }

            if (policy == null) {
                mAdFirstInsertValid = true;
                mPolicy = null;
                logHelper.infoShow(" setAdvertisementPolicy is NULL (" + mAdFirstInsertValid + ")");
                return;
            }

            logHelper.infoShow(" policy-setpolicy first validtime Maxsongs volidtimes(" + policy.adpFirstInsertSongNum
                    + "," + policy.adpSingalSongValidTime + "," + policy.adpSingalSongMaxCount + ","
                    + policy.pMaxTotalTime + "," + mAdFirstInsertValid + "," + mIsInsertImmediatelyOnce + ")");

            if (policy.adpFirstInsertSongNum < 2) {
                mAdFirstInsertValid = false;
            }

            mPolicy = new AdvertiseMentPolicyInline();
            if (mPolicy == null)
                return;

            if (policy.adpSingalSongValidTime == 0) {
                policy.adpSingalSongValidTime = 60;// default 60s ;
            }

            if (0 == policy.pMaxTotalTime) {
                mIsInsertImmediately = true;
                policy.pMaxTotalTime = 60;
            } else {
                mIsInsertImmediately = false;
            }

            mPolicy.bNeedCalcAdProgress = policy.bNeedCalcAdProgress;
            mPolicy.pFirstInsertSongNum = policy.adpFirstInsertSongNum; // first
                                                                        // insert
                                                                        // after
                                                                        // songnum
                                                                        // ;
            mPolicy.pSingalSongValidTime = policy.adpSingalSongValidTime; // the
                                                                          // time
                                                                          // of
                                                                          // singal
                                                                          // song
                                                                          // is
                                                                          // valid;
            mPolicy.pSingalSongMaxCount = policy.adpSingalSongMaxCount; // first
                                                                        // condition
                                                                        // ; the
                                                                        // number
                                                                        // of
                                                                        // valid
                                                                        // song;
            mPolicy.pMaxTotalTime = policy.pMaxTotalTime; // second condition ;
                                                          // total time of
                                                          // valid playing ;

            // 0和1意义相同
            if (1 == mPolicy.pFirstInsertSongNum) {
                mPolicy.pFirstInsertSongNum = 0;
            }

            mAdTimerRelativeSize = 0;

            logHelper.infoShow(" policy-checkpolicy first validtime Maxsongs volidtimes(" + mPolicy.pFirstInsertSongNum
                    + "," + mPolicy.pSingalSongValidTime + "," + mPolicy.pSingalSongMaxCount + ","
                    + mPolicy.pMaxTotalTime + "," + mAdFirstInsertValid + "," + mIsInsertImmediatelyOnce + ")");

            // MUSIC
            mPolicy.pFirstInsertSongFinished = false;
            mPolicy.pSingalSongMaxCountCalcValid = true;

            mPolicy.pSingalSongValidSize = 0;

            mPolicy.pSingalSongValidCount = 0;
            mPolicy.pSingalSongSumSize = 0;
            mPolicy.pSingalSongSumLastSize = 0;
            // 广告开始播放后设置策略不应该清除广播放状态，统一在reset中作了广告播放状态设置
            // mPolicy.pAdPlayingValid = false;
            mPolicy.pSingalAdSumSize = 0;
            mPolicy.pMultiAdSumSize = 0;

            if ((0 == mPolicy.pFirstInsertSongNum) && (0 == mPolicy.pSingalSongMaxCount)) {// 第一个音频开始符合条件的广告就可以插入
                mAdInsertConditionSingalSongMaxCountIsValid = true;
                mAdTimerRelativeMaxTime = mPolicy.pMaxTotalTime;
            } else {
                mAdInsertConditionSingalSongMaxCountIsValid = false;
                mAdTimerRelativeMaxTime = 0;
            }

            // 设置策略的时候没有正在播放广告则重置本次场景是否已经播放广告状态
            if (!mPolicy.pAdPlayingValid) {
                mAdhavePlayedValid = false;
            }
            mAdSingalPreDownloadValid = true;
            mAdSingalDownloadValid = true;
            mAdSingalInsertValid = true;
            mAdWillPlayValid = false;
        }

        // InitAdParm if setAdvertisementPolicy is called when mIsPrepared
        if (mIsPrepared) {
            InitAdParm();
        }
    }

    /**
     * bypass current ad playing, continue to play main audio from current insert position, the caller will receive
     * AD_INTERRUPT *
     * 
     * @throws IllegalStateException if it is called in an invalid state
     */

    public void bypassCurrentAd() throws IllegalStateException {
        logHelper.infoShow(" policy BypassCurrentAd ");
        if ((mPolicy != null) && (mPolicy.pAdPlayingValid)) {
            logHelper.infoShow(" policy BypassCurrentAd ad is playing, mAdCurAnchorIndex " + mAdCurAnchorIndex);
            if ((mAdCurAnchorIndex >= 0) && (mAdCurAnchorIndex < mAdInfos.anchorNums) && (mAdInfos.anchorTime != null)) {
                seekTo(mAdInfos.anchorTime[mAdCurAnchorIndex]);
            }
        }
    }

    /**
     * Prepares the player for playback, synchronously.
     * 
     * After setting the datasource and the display surface, you need to either call prepare() or prepareAsync(). For
     * files, it is OK to call prepare(), which blocks until AudioPlayer is ready for playback.
     * 
     * @throws IllegalStateException if it is called in an invalid state
     */
    public void prepare() throws IOException, IllegalStateException {
        mIsPrepareAsync = false;
        _prepare();

        mStartPos = 0;
        // mStartSamples = 0;
        mAudioTrackSum = 0;
        mAudioTrackSumStart = 0;
        mAudioTrackSumOriginal = 0;
        mAudioTrackSumLast = 0;
        mDuration = 0;
        mFadeReferentCount = 0;
        mIsPlaying = false;

        // sync. call onPreparedSuccessInner
        if (onPreparedSuccessInner()) {
            mIsPrepared = true;
        }
    }

    /**
     * Prepares the player for playback, asynchronously.
     * 
     * After setting the datasource and the display surface, you need to either call prepare() or prepareAsync(). For
     * streams, you should call prepareAsync(), which returns immediately, rather than blocking until enough data has
     * been buffered.
     * 
     * @throws IllegalStateException if it is called in an invalid state
     */
    public void prepareAsync() throws IOException, IllegalStateException {
        mIsPrepareAsync = true;
        _prepareAsync();

        mStartPos = 0;
        // mStartSamples = 0;
        mAudioTrackSum = 0;
        mAudioTrackSumStart = 0;
        mAudioTrackSumOriginal = 0;
        mAudioTrackSumLast = 0;
        mDuration = 0;
        mFadeReferentCount = 0;
        mIsPlaying = false;
    }

    protected void initAudiotrack() throws IllegalStateException {
        // call _setVisualBuffer before _start()
        // set visual buffer
        // logHelper.infoShow("audioplayer.java _setVisualBuffer!")

        int sampleRateInHz = this.getRenderSampleRate();
        mSampleRateInHz = sampleRateInHz;
        int channelConfig = this.getRenderChannels();
        int audioFormat = this.getRenderBitsPerSample();
        int bytepersec = sampleRateInHz * channelConfig * audioFormat / 8;
        mBytePerSec = bytepersec;

        if (mOnGetRenderFormatListener != null) {
            mOnGetRenderFormatListener.onGetRenderFormat(sampleRateInHz, channelConfig, audioFormat, this);
        }

        if (channelConfig == 1) {
            channelConfig = AudioFormat.CHANNEL_OUT_MONO;
        } else {
            channelConfig = AudioFormat.CHANNEL_OUT_STEREO;
        }
        // Only support 16bit.
        audioFormat = AudioFormat.ENCODING_PCM_16BIT;

        int minBufSize = AudioTrack.getMinBufferSize(sampleRateInHz, channelConfig, audioFormat);
        if (minBufSize <= 0)
            throw new IllegalStateException("minBufSize less than zero");
        logHelper.infoShow("minBufSize = " + minBufSize);
        // set minBufSize to visualbuffer, minBufSize*2 audio sync why?
        _setVisualBuffer(minBufSize*2);

        AudioTrack audioTrack = mAudioTrack;
        if (audioTrack == null) {
            synchronized (this) {
                // max buffer is 0.5 sec

                int maxBufSize = minBufSize > bytepersec / 2 ? minBufSize : bytepersec / 2;
                minBufSize = minBufSize * 8 > maxBufSize ? maxBufSize : minBufSize * 8;
                logHelper.infoShow("bytepersec = " + bytepersec);
                logHelper.infoShow("maxBufSize = " + maxBufSize);
                logHelper.infoShow("minBufSize = " + minBufSize);
                minBufSize = (minBufSize / 1024 + 1) * 1024;
                logHelper.infoShow("minBufSize = " + minBufSize);

                /*
                 * if(android.os.Build.VERSION.SDK_INT >= 9){ mAudioTrack = new AudioTrack(mStreamtype, sampleRateInHz,
                 * channelConfig, audioFormat, minBufSize, AudioTrack.MODE_STREAM,mSessionID); }else{ mAudioTrack = new
                 * AudioTrack(mStreamtype, sampleRateInHz, channelConfig, audioFormat, minBufSize,
                 * AudioTrack.MODE_STREAM); }
                 */
                if (mSessionID != 0)// if setAudioSessionId is called, construct
                                    // audiotrack with mSessionID
                {
                    try {
                        Class<?> classType = Class.forName("android.media.AudioTrack");
                        Constructor<?> constructor =
                                classType.getDeclaredConstructor(int.class, int.class, int.class, int.class, int.class,
                                        int.class, int.class); // find construct
                                                               // with
                                                               // sessionid
                        mAudioTrack =
                                (AudioTrack) constructor.newInstance(mStreamtype, sampleRateInHz, channelConfig,
                                        audioFormat, minBufSize, AudioTrack.MODE_STREAM, mSessionID);
                    } catch (NoSuchMethodException ex) {
                        ex.printStackTrace();
                    } catch (Exception ex) {
                        ex.printStackTrace();
                    }
                }

                if (mAudioTrack == null) // if construct with sessionid not
                                         // available, use old constructor
                {
                    mAudioTrack =
                            new AudioTrack(mStreamtype, sampleRateInHz, channelConfig, audioFormat, minBufSize,
                                    AudioTrack.MODE_STREAM);
                }

                _setSpeed();
                mBufferSize = minBufSize;
                audioTrack = mAudioTrack;
                // mStartPos = 0;
                // mStartSamples = 0;
                mAudioTrackSumStart = 0;
                mAudioTrackSumOriginal = 0;
            }

            setVolumeSys(mVolumeLeft, mVolumeRight, null);
        }
    }

    /**
     * Starts or resumes playback. If playback had previously been paused, playback will continue from where it was
     * paused. If playback had been stopped, or never started before, playback will start at the beginning.
     * 
     * @throws IllegalStateException if it is called in an invalid state
     */
    public void start() throws IllegalStateException {
        logHelper.infoShow("start start ");

        if (!mIsPrepared) {
            logHelper.infoShow("start error not prepared ");
            return;
        }
        stayAwake(true);
        // call _setVisualBuffer before _start()
        // set visual buffer
        // logHelper.infoShow("audioplayer.java _setVisualBuffer!")

        _start();

        logHelper.infoShow("start start outter");
        AudioTrack audioTrack = mAudioTrack;
        if (audioTrack != null) {
            startFadeInFadeOut(-1, 0, FADE_MODE_PLAY);// fade in
            logHelper.infoShow("audioTrack play begin ");

            try {
                int nPlayStatus = audioTrack.getPlayState();
                if (AudioTrack.PLAYSTATE_PLAYING != nPlayStatus) {
                    logHelper.infoShow("audioTrack play call play ");
                    audioTrack.play();
                } else {
                    logHelper.infoShow("audioTrack play not call play ");
                }
            } catch (Exception e) {
                e.printStackTrace();
            }

            logHelper.infoShow("audioTrack play end ");
        }

        if (mRenderhandler != null) {
            mRenderhandler.removeMessages(RENDER_START);
            mRenderhandler.sendEmptyMessage(RENDER_START);
        }

        mIsPlaying = true;
        logHelper.infoShow("start end ");
    }

    /**
     * Stops playback after playback has been stopped or paused.
     * 
     * @throws IllegalStateException if the internal player engine has not been initialized.
     */
    public void stop() throws IllegalStateException {
        logHelper.infoShow("stop start ");
        // stayAwake(false);
        // releaseAudiotrack();
        // _stop();
        //
        // mIsPlaying = false;
        // do not use close, just reset
        reset();
        logHelper.infoShow("stop end ");
    }

    /**
     * Pauses playback. Call start() to resume.
     * 
     * @throws IllegalStateException if the internal player engine has not been initialized.
     */
    public void pause() throws IllegalStateException {
        logHelper.infoShow("pause start ");
        if (!mIsPrepared) {
            logHelper.infoShow("pause error not prepared ");
            return;
        }
        stayAwake(false);
        if (mRenderhandler != null) {
            mRenderhandler.removeMessages(RENDER_STOP);
            mRenderhandler.sendEmptyMessage(RENDER_STOP);
        }

        if (!mIsEnableFadeInFadeOut || (mFadeReferentCount != 0)) {
        	logHelper.infoShow("pause mIsEnableFadeInFadeOut:" + mIsEnableFadeInFadeOut + " mFadeReferentCount:" + mFadeReferentCount);
            try {
                AudioTrack audioTrack = mAudioTrack;
                if (audioTrack != null) {
                    audioTrack.pause();
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        } else {
            startFadeInFadeOut(0, PAUSE_FADEIOUT_SLEEP_MS, FADE_MODE_PAUSE);// fade out
        }

        _pause();

        mIsPlaying = false;
        logHelper.infoShow("pause end ");
    }

    /**
     * Set the low-level power management behavior for this AudioPlayer.
     * 
     * <p>
     * This function has the AudioPlayer access the low-level power manager service to control the device's power usage
     * while playing is occurring. The parameter is a combination of {@link android.os.PowerManager} wake flags. Use of
     * this method requires {@link android.Manifest.permission#WAKE_LOCK} permission. By default, no attempt is made to
     * keep the device awake during playback.
     * 
     * @param context the Context to use
     * @param mode the power/wake mode to set
     * @see android.os.PowerManager
     */
    public void setWakeMode(Context context, int mode) {
        logHelper.infoShow("setWakeMode start ");
        boolean washeld = false;
        if (mWakeLock != null) {
            if (mWakeLock.isHeld()) {
                washeld = true;
                mWakeLock.release();
            }
            mWakeLock = null;
        }

        PowerManager pm = (PowerManager) context.getSystemService(Context.POWER_SERVICE);
        mWakeLock = pm.newWakeLock(mode | PowerManager.ON_AFTER_RELEASE, AudioPlayer.class.getName());
        mWakeLock.setReferenceCounted(false);
        if (washeld) {
            mWakeLock.acquire();
        }
        logHelper.infoShow("setWakeMode end ");
    }

    private void stayAwake(boolean awake) {
        try {
            logHelper.infoShow("stayAwake start ");
            if (mWakeLock != null) {
                if (awake && !mWakeLock.isHeld()) {
                    mWakeLock.acquire();
                } else if (!awake && mWakeLock.isHeld()) {
                    mWakeLock.release();
                }
            }
            mStayAwake = awake;
            logHelper.infoShow("stayAwake end ");
            return;
        } catch (Exception ex) {
            logHelper.errorShow(ex.toString());
        }
    }

    /**
     * Checks whether the AudioPlayer is playing.
     * 
     * @return true if currently playing, false otherwise
     */

    public boolean isPlaying() {
        boolean isPlaying = false;
        if (mAudioTrack != null) {
            isPlaying = mIsPlaying;
        }
        return isPlaying;
    }

    /**
     * Seeks to specified time position.
     * 
     * @param msec the offset in milliseconds from the start to seek to
     * @throws IllegalStateException if the internal player engine has not been initialized
     */
    public void seekTo(int msec) throws IllegalStateException {
        logHelper.infoShow("mlcheck_seekTo start " + msec);

        if (!mIsPrepared) {
            logHelper.infoShow("seekTo error not prepared ");
            return;
        }

        // 拖动进度时判断是否正在播放广告，如果正在播放和播放广告时切歌逻辑相同
        checkAdInterrupt();

        int msecLast = mStartPosTem;
        if (msec >= 0) {
          //  mStartPosTem = msec;
            mStartPos = msec;
            mAudioTrackSumOriginal = mAudioTrackSumStart;
            mAudioTrackSumStart = mAudioTrackSum;
            logHelper.infoShow("mlcheck_seekTo mAudioTrackSum " + mAudioTrackSum+" msecLast= "+msecLast);
            try {
                _seekTo(msec);
            } catch (IllegalStateException e) {// when error, reset mStartPosTem
                                               // to old value
                logHelper.errorShow("seekTo error mStartPosTem" + mStartPosTem + "msecLast" + msecLast);
                mStartPosTem = msecLast;
                throw e;
            }
            // mStartPos = msec;
        }

        logHelper.infoShow("seekTo end ");
    }

    /**
     * Gets the advertisement duration
     * 
     * @param advertisement songid
     * @return the ad duration
     */
    public int getAdvertiseMentDuration(int adId) {
        int i = 0;
        int duration = 0;
        if ((mAdInfos == null) || (mAdInfos.adId == null)) {
            logHelper.infoShow("getCurrentAdvertiseMentDuration Failed\n");
            return -1;
        }
        for (i = 0; i < mAdInfos.adNums; i++) {
            if (mAdInfos.adId[i] == adId) {
                duration = mAdInfos.adDuration[i];
            }
        }
        return duration;
    }

    /**
     * Gets the current playback position of insert audio eg.advertisement.
     * 
     * @return the current position in milliseconds
     */
    public int getCurrentAdPosition() {
        int position = 0;
        if (!mIsPrepared) {
            logHelper.infoShow("getCurrentAdPosition error not prepared ");
            return 0;
        }

        if (mBytePerSec == 0) {
            return 0;
        }

        AudioTrack audioTrack = mAudioTrack;
        try {
            if (audioTrack != null) {
                if ((mPolicy != null) && (mPolicy.pAdPlayingValid)) {
                    position = (int) ((mPolicy.pMultiAdSumSize * 1000) / mBytePerSec);
                } else {
                    position = 0;
                }
            }
        } catch (Exception ex) {
            logHelper.errorShow(ex.toString());
        }
        if (position <= 0)
            position = 1;

        return position;
    }

    /**
     * Gets the current playback position.
     * 
     * @return the current position in milliseconds
     */

    public int getCurrentPosition() {
        int position = 0;
        if (!mIsPrepared) {
            // logHelper.infoShow("getCurrentPosition error not prepared ");
            return 0;
        }
        if (mBytePerSec == 0) {
            return 0;
        }
        AudioTrack audioTrack = mAudioTrack;
        try {
            if (audioTrack != null) {
                // int playRate = audioTrack.getPlaybackRate();
                // int playRate = mSampleRateInHz;
                // if (playRate <= 0)
                // playRate = 44100;
                // ZTE blocks when getPlaybackHeadPosition
                // position = mStartPos
                // + (int) (((long) (audioTrack.getPlaybackHeadPosition() -
                // mStartSamples) * 1000) / playRate);
                // logHelper.infoShow("getCurrentPosition api:" + position);

                if ((mPolicy != null) && (mPolicy.bNeedCalcAdProgress) && (mPolicy.pAdPlayingValid)) {
                    position = (int) ((mPolicy.pMultiAdSumSize * 1000) / mBytePerSec);
                } else {
                    long lAudioTrackByte = mAudioTrackSum - mAudioTrackSumStart;
                    if (lAudioTrackByte >= mBufferSize)// delay between write
                                                       // and
                                                       // sound
                        lAudioTrackByte -= mBufferSize;
                    
                    position = mStartPos + (int) ((lAudioTrackByte * 1000) / mBytePerSec);
                }
              //   logHelper.infoShow("mlcheck_pos="+position+",tracksum="+mAudioTrackSum+",bytepersec="+mBytePerSec+",startpos="+mStartPos+",sumstart="+mAudioTrackSumStart+")");
              //   logHelper.infoShow("getCurrentPosition sum:" + position);

            }
        } catch (Exception ex) {
            logHelper.errorShow(ex.toString());
        }
        if (position <= 0)
            position = 1;

        // position must less than duration
        int nDuration = getDuration();
        if ((nDuration > 0) && (position > nDuration)) {
        	logHelper.warnShow("getCurrentPosition position " + position + " nDuration " + nDuration);
            position = nDuration;
        }

        return position;
    }

    /**
     * Sets the refrence duration of the file, 0 will be ignored.should be called before prepare/prepareasync
     * 
     * @param durationMillisec the duration in milliseconds
     */
    public void setRefrenceDuration(int durationMillisec) {
        native_setRefrenceDuration(durationMillisec);
    }

    /**
     * Gets the duration of the file.
     * 
     * @return the duration in milliseconds
     */
    public int getDuration() {
        if (!mIsPrepared) {
            // logHelper.infoShow("getDuration error not prepared ");
            return 0;
        }

        if (mDuration == 0) {
            mDuration = native_getDuration();
        }

        return mDuration;
    }

    /**
     * Releases resources associated with this AudioPlayer object. It is considered good practice to call this method
     * when you're done using the AudioPlayer.
     */
    public void release() {
        reset();
        logHelper.infoShow("release start ");
        stayAwake(false);
        // releaseAudiotrack();//called in reset
        if (mRenderThread != null) {
            mRenderThread.getLooper().quit();
        }
        mRenderThread = null;
        mRenderhandler = null;

        mOnAdvertiseMentPlayStatusListener = null;
        mOnPreparedListener = null;
        mOnBufferingUpdateListener = null;
        mOnCompletionListener = null;
        mOnSeekCompleteListener = null;
        mOnErrorListener = null;
        mOnInfoListener = null;

        mAdInfos = null;
        mAdReport = null;

        _release();
        logHelper.infoShow("release end ");
    }

    private void _stopRenderHandler() {
        if (mRenderhandler != null) {
            // logHelper.infoShow("begin removeMessages RENDER_STOP!");
            mRenderhandler.removeMessages(RENDER_STOP);
            // logHelper.infoShow("begin sendEmptyMessage RENDER_STOP!");
            mRenderhandler.sendEmptyMessage(RENDER_STOP);
        }
    }

    private void _releaseAudiotrack() {
        synchronized (this) {
            AudioTrack audioTrack = mAudioTrack;
            try {
                if (audioTrack != null) {
                    audioTrack.pause();
                    audioTrack.flush();
                    audioTrack.stop();
                    audioTrack.release();
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
            mAudioTrack = null;
        }
    }

    private void releaseAudiotrack() {
        _stopRenderHandler();
        _releaseAudiotrack();
    }

    /**
     * Resets the AudioPlayer to its uninitialized state. After calling this method, you will have to initialize it
     * again by setting the data source and calling prepare().
     */
    public void reset() {
        logHelper.infoShow("reset start ");
        stayAwake(false);

        _stopRenderHandler();

//        if (mIsPlaying) {
//            startFadeInFadeOut(0, -1, FADE_MODE_STOP);
//        }
        _reset();
        // make sure none of the listeners get called anymore
        if (mEventHandler != null) {
            mEventHandler.removeCallbacksAndMessages(null);
        }

        if (advertiseMentHandler != null) {
            advertiseMentHandler.removeCallbacksAndMessages(null);
        }

        mStartPos = 0;
        // mStartSamples = 0;
        mAudioTrackSumStart = 0;
        mAudioTrackSumOriginal = 0;

        mIsPlaying = false;
        mIsPrepared = false;
        mDuration = 0;

        _releaseAudiotrack();

        checkAdInterrupt();

        if (mTaskIdList != null) {
            for (int i = 0; i < mTaskIdList.size(); ++i) {
                deleteDownloadTask(mTaskIdList.get(i));
            }
            mTaskIdList.clear();
        }
        logHelper.infoShow("reset end ");
    }

    /**
     * Sets the audio stream type for this AudioPlayer. See {@link AudioManager} for a list of stream types. Must call
     * this method before prepare() or prepareAsync() in order for the target stream type to become effective
     * thereafter.
     * 
     * @param streamtype the audio stream type
     * @see com.baidu.util.audiocore.AudioManager
     */
    public void setAudioStreamType(int streamtype) {
        mStreamtype = streamtype;
    }

    /**
     * /** Returns the audio session ID.
     * 
     * @return the audio session ID. Note that the audio session ID is 0 only if a problem occured when the AudioPlayer
     *         was contructed. Only support API Level 9 or higher. Only useful after call start().
     */
    public int getAudioSessionId() {
        int sessionId = 0;

        /*
         * if(mAudioTrack != null && android.os.Build.VERSION.SDK_INT >= 9) sessionId = mAudioTrack.getAudioSessionId();
         */
        if (mAudioTrack != null) {
            Method method = null;
            try {
                Class<? extends AudioTrack> classType = mAudioTrack.getClass();
                method = classType.getDeclaredMethod("getAudioSessionId", new Class[] {});
                sessionId = (Integer) method.invoke(mAudioTrack, new Object[] {});

            } catch (NoSuchMethodException ex) {
                ex.printStackTrace();
            } catch (Exception ex) {
                ex.printStackTrace();
            }

            // sessionId = mAudioTrack.getAudioSessionId();
        }

        return sessionId;
    }

    /**
     * Sets the audio session ID.
     * 
     * @param sessionId the audio session ID. The audio session ID is a system wide unique identifier for the audio
     *            stream played by this audioplayer instance. The primary use of the audio session ID is to associate
     *            audio effects to a particular instance of MediaPlayer: if an audio session ID is provided when
     *            creating an audio effect, this effect will be applied only to the audio content of media players
     *            within the same audio session and not to the output mix. When created, a audioplayer instance
     *            automatically generates its own audio session ID. However, it is possible to force this player to be
     *            part of an already existing audio session by calling this method. This method must be called before
     *            start() methods.
     */
    public void setAudioSessionId(int sessionId) {
        mSessionID = sessionId;
    }

    // modify
    public static final int AD_UNKNOWN = 0x00;
    public static final int AD_WILLPLAY = 0x01;
    public static final int AD_PREDOWNLOAD = 0x02;
    public static final int AD_DOWNLOAD = 0x03;
    public static final int AD_INSERT = 0x04;
    public static final int AD_PLAYING = 0x05;
    public static final int AD_FINISHED = 0x06;
    public static final int AD_ALLFINISHED = 0x07;
    public static final int AD_INTERRUPT = 0x08;
    public static final int AD_MAX = 0x08;

    public interface OnAdvertiseMentPlayStatusListener {
        /**
         * Called when the status of ad is changed. status 0x01 PLAYING; 0x02 FINISHED; 0x04 ALLFINISHED; 0x08
         * AD_INTERRUPT
         * 
         * @param status 0x01 PLAYING; 0x02 FINISHED; 0x04 ALLFINISHED; 0x08 AD_INTERRUPT
         * @param id the ID of AD
         * @param pos cur play pos(millisec) of ad, used when AD_INTERRUPT
         */
        void onAdvertiseMentPlayStatus(AudioPlayer ap, int status, int id, int pos);
    }

    public void setOnAdvertiseMentPlayStatusListener(OnAdvertiseMentPlayStatusListener listener) {
        mOnAdvertiseMentPlayStatusListener = listener;
    }

    private OnAdvertiseMentPlayStatusListener mOnAdvertiseMentPlayStatusListener;

    /**
     * Interface definition for a callback to be invoked when the media source is ready for playback.
     */
    public interface OnPreparedListener {
        /**
         * Called when the media file is ready for playback.
         * 
         * @param ap the AudioPlayer that is ready for playback
         */
        void onPrepared(AudioPlayer ap);
    }

    /**
     * Register a callback to be invoked when the media source is ready for playback.
     * 
     * @param listener the callback that will be run
     */
    public void setOnPreparedListener(OnPreparedListener listener) {
        mOnPreparedListener = listener;
    }

    private OnPreparedListener mOnPreparedListener;

    /**
     * Interface definition for a callback to be invoked when playback of a media source has completed.
     */
    public interface OnCompletionListener {
        /**
         * Called when the end of a media source is reached during playback.
         * 
         * @param ap the AudioPlayer that reached the end of the file
         */
        void onCompletion(AudioPlayer ap);
    }

    /**
     * Register a callback to be invoked when the end of a media source has been reached during playback.
     * 
     * @param listener the callback that will be run
     */
    public void setOnCompletionListener(OnCompletionListener listener) {
        mOnCompletionListener = listener;
    }

    private OnCompletionListener mOnCompletionListener;

    /**
     * percent from buffer start to buffer end Interface definition of a callback to be invoked indicating buffering
     * status of a media resource being streamed over the network.
     */
    public interface OnBufferingUpdateListener {
        /**
         * Called to update status in buffering a media stream.
         * 
         * @param ap the AudioPlayer the update pertains to
         * @param percent the percentage (0-100) of the buffer that has been filled thus far
         */
        void onBufferingUpdate(AudioPlayer ap, int percent);
    }

    /**
     * Register a callback to be invoked when the status of a network stream's buffer has changed.
     * 
     * @param listener the callback that will be run.
     */
    public void setOnBufferingUpdateListener(OnBufferingUpdateListener listener) {
        mOnBufferingUpdateListener = listener;
    }

    private OnBufferingUpdateListener mOnBufferingUpdateListener;

    /**
     * Interface definition of a callback to be invoked indicating the completion of a seek operation.
     */
    public interface OnSeekCompleteListener {
        /**
         * Called to indicate the completion of a seek operation.
         * 
         * @param ap the AudioPlayer that is ready for playback is succeed, null if prepare is failed
         */
        public void onSeekComplete(AudioPlayer ap);
    }

    /**
     * Register a callback to be invoked when a seek operation has been completed.
     * 
     * @param listener the callback that will be run
     */
    public void setOnSeekCompleteListener(OnSeekCompleteListener listener) {
        mOnSeekCompleteListener = listener;
    }

    private OnSeekCompleteListener mOnSeekCompleteListener;

    /**
     * download bytes from http and p2p，callback is called when download task complete .
     */
    public interface OnDownloadBytesListener {
        /**
         * Called to update status in buffering a media stream.
         * 
         * @param downloadbytes download bytes from http and p2p this task
         */
        void onDownloadBytes(int downloadbytes);
    }

    /**
     * Register a callback to be invoked when download task is end.
     * 
     * @param listener the callback that will be run.
     */
    public static void setOnDownloadBytesListener(OnDownloadBytesListener listener) {
        mOnDownloadBytesListener = listener;
        native_enableDownloadBytesCallback();
    }

    private static OnDownloadBytesListener mOnDownloadBytesListener = null;

    /**
     * Get DNS listener.OnSuccess if succcess, OnError is failed
     */
    public interface OnHttpResponseListener {
        /**
         * success
         * 
         * @param url: url of http request
         * @param code: code of http response
         * @param text: body of http response
         */
        void onHttpResponse(String url, int code, String text);
    }

    public static void httpRequestFromUrl(String url, int retrytime, int timeout, HashMap<String, String> headerMap, 
            OnHttpResponseListener listener) throws IllegalArgumentException {
        mOnHttpResponseListener = listener;
        
        String headers[] = new String[headerMap.size()];

        int i = 0;
        Iterator<String> iter = headerMap.keySet().iterator();

        while (iter.hasNext()) {
            headers[i] = (String) iter.next();
            i++;
        }

        i = 0;
        iter = headerMap.values().iterator();
        while (iter.hasNext()) {
            headers[i] = headers[i] + ":" + (String) iter.next();
            i++;
        }
        native_httpRequest(url, retrytime, timeout, headers);
    }

    private static OnHttpResponseListener mOnHttpResponseListener = null;
    /*
     * Do not change these values without updating their counterparts in include/media/mediaplayer.h!
     */
    /**
     * Unspecified media player error.
     * 
     * @see com.baidu.util.audiocore.AudioPlayer.OnErrorListener
     */
    // public static final int MEDIA_ERROR_OK = 0; //没有错误
    public static final int MEDIA_ERROR_UNKNOWN = 1; // 未知错误
    public static final int MEDIA_ERROR_SDCARDSPACE = 2; // 空间不足
    public static final int MEDIA_ERROR_AUDIOTRACK = 3; // 播放系统错误
    public static final int MEDIA_ERROR_NET_TIMEOUT = 10; // 网络超时
    public static final int MEDIA_ERROR_NET_CREATEFILE = 11; // 创建缓存文件错误
    public static final int MEDIA_ERROR_NET_WRITEFILE = 12; // 缓存文件写入错误
    public static final int MEDIA_ERROR_NET_CONNECT = 13; // 网络连接错误
    public static final int MEDIA_ERROR_NET_OTHER = 14; // 其他网络错误
    public static final int MEDIA_ERROR_FILE_SIZE = 20; // 文件大小错误
    public static final int MEDIA_ERROR_AUDIO_PARSE = 30; // 音频格式解析错误
    public static final int MEDIA_ERROR_AUDIO_CODEC = 31; // 音频解码错误
    public static final int MEDIA_ERROR_SEEK = 40; // 拖动进度错误

    /**
     * Interface definition of a callback to be invoked when there has been an error during an asynchronous operation
     * (other errors will throw exceptions at method call time).
     */
    public interface OnErrorListener {
        /**
         * Called to indicate an error.
         * 
         * @param ap the AudioPlayer the error pertains to
         * @param what the type of error that has occurred:
         *            <ul>
         *            <li>{@link #MEDIA_ERROR_UNKNOWN}
         *            <li>{@link #MEDIA_ERROR_SDCARDSPACE} ...
         *            </ul>
         * @param extra an extra code, specific to the error. Typically implementation dependant.
         * @return True if the method handled the error, false if it didn't. Returning false, or not having an
         *         OnErrorListener at all, will cause the OnCompletionListener to be called.
         */
        boolean onError(AudioPlayer ap, int what, int extra);
    }

    /**
     * Register a callback to be invoked when an error has happened during an asynchronous operation.
     * 
     * @param listener the callback that will be run
     */
    public void setOnErrorListener(OnErrorListener listener) {
        mOnErrorListener = listener;
    }

    private OnErrorListener mOnErrorListener;

    /*
     * Do not change these values without updating their counterparts in include/media/mediaplayer.h!
     */
    /**
     * not used. Unspecified media player info.
     * 
     * @see com.baidu.util.audiocore.AudioPlayer.OnInfoListener
     */
    public static final int MEDIA_INFO_UNKNOWN = 1;

    /**
     * not used. The video is too complex for the decoder: it can't decode frames fast enough. Possibly only the audio
     * plays fine at this stage.
     * 
     * @see com.baidu.util.audiocore.AudioPlayer.OnInfoListener
     */
    public static final int MEDIA_INFO_VIDEO_TRACK_LAGGING = 700;

    /**
     * net buffering is slower than need, begin buffering. At this time, render buffer is not empty, so render is still
     * working.
     */
    public static final int MEDIA_INFO_BUFFERING_START = 701;

    /**
     * AudioPlayer is resuming playback after filling buffers.
     */
    public static final int MEDIA_INFO_BUFFERING_END = 702;

    /**
     * buffer is slower than play, not caused by user's operation.At this time, render buffer is empty, render is
     * paused.
     */
    public static final int MEDIA_INFO_BUFFERING_AUTO = 704;

    /**
     * not used. Bad interleaving means that a media has been improperly interleaved or not interleaved at all, e.g has
     * all the video samples first then all the audio ones. Video is playing but a lot of disk seeks may be happening.
     * 
     * @see com.baidu.util.audiocore.AudioPlayer.OnInfoListener
     */
    public static final int MEDIA_INFO_BAD_INTERLEAVING = 800;

    /**
     * not used. The media cannot be seeked (e.g live stream)
     * 
     * @see com.baidu.util.audiocore.AudioPlayer.OnInfoListener
     */
    public static final int MEDIA_INFO_NOT_SEEKABLE = 801;

    /**
     * not used. A new set of metadata is available.
     * 
     * @see com.baidu.util.audiocore.AudioPlayer.OnInfoListener
     */
    public static final int MEDIA_INFO_METADATA_UPDATE = 802;

    /**
     * audioplayer is begin to play.
     * 
     * @see com.baidu.util.audiocore.AudioPlayer.OnInfoListener
     */
    public static final int MEDIA_INFO_RENDER_START = 900;

    /**
     * not used. audioplayer is end to play.
     * 
     * @see com.baidu.util.audiocore.AudioPlayer.OnInfoListener
     */
    public static final int MEDIA_INFO_RENDER_END = 901;

    /**
     * audioplayer is blocked when play. used for statistics
     * 
     * @see com.baidu.util.audiocore.AudioPlayer.OnInfoListener
     */
    public static final int MEDIA_INFO_NEEDMOREDATA = 902;

    /**
     * Interface definition of a callback to be invoked to communicate some info and/or warning about the media or its
     * playback.
     */
    public interface OnInfoListener {
        /**
         * Called to indicate an info or a warning.
         * 
         * @param ap the AudioPlayer the info pertains to.
         * @param what the type of info or warning.
         *            <ul>
         *            <li>{@link #MEDIA_INFO_UNKNOWN}
         *            <li>{@link #MEDIA_INFO_VIDEO_TRACK_LAGGING}
         *            <li>{@link #MEDIA_INFO_BAD_INTERLEAVING}
         *            <li>{@link #MEDIA_INFO_NOT_SEEKABLE}
         *            <li>{@link #MEDIA_INFO_METADATA_UPDATE}
         *            </ul>
         * @param extra an extra code, specific to the info. Typically implementation dependant.
         * @return True if the method handled the info, false if it didn't. Returning false, or not having an
         *         OnErrorListener at all, will cause the info to be discarded.
         */
        boolean onInfo(AudioPlayer ap, int what, int extra);
    }

    /**
     * Register a callback to be invoked when an info/warning is available.
     * 
     * @param listener the callback that will be run
     */
    public void setOnInfoListener(OnInfoListener listener) {
        mOnInfoListener = listener;
    }

    private OnInfoListener mOnInfoListener;

    /**
     * Interface definition for a callback to be invoked when player has got PCM data from native during playing. For
     * LBS carlife
     */
    public interface OnGetRenderBufferListener {
        /**
         * Called when player has got PCM data from native during playing. make sure OnGetRenderBuffer return
         * immediately, or playback may be blocked
         * 
         * @param bsize size of buffer in byte
         * @param buffer PCM data
         * @param ap the AudioPlayer that is playing
         */
        void onGetRenderBuffer(int bsize, byte[] buffer, AudioPlayer ap);
    }

    /**
     * Register a callback to be invoked when player has got PCM data from native during playing. For LBS carlife
     * 
     * @param listener the callback that will be run
     */
    public void setOnGetRenderBufferListener(OnGetRenderBufferListener listener) {
        mOnGetRendderBufferListener = listener;
    }

    private OnGetRenderBufferListener mOnGetRendderBufferListener = null;

    /**
     * Interface definition for a callback to be invoked when player has got PCM data from native during playing.
     */
    public interface OnGetRenderFormatListener {
        /**
         * Called when player has got PCM data from native during playing. make sure OnGetRenderBuffer return
         * immediately, or playback may be blocked
         * 
         * @param samplerate samplerate of render output PCM
         * @param channels channels of render output PCM
         * @param bitspersample bitspersample of render output PCM
         * @param ap the AudioPlayer that is playing
         */
        void onGetRenderFormat(int samplerate, int channels, int bitspersample, AudioPlayer ap);
    }

    /**
     * Register a callback to be invoked when player has got PCM data from native during playing.
     * 
     * @param listener the callback that will be run
     */
    public void setOnGetRenderFormatListener(OnGetRenderFormatListener listener) {
        mOnGetRenderFormatListener = listener;
    }

    private OnGetRenderFormatListener mOnGetRenderFormatListener = null;

    // effect index These constants must be synchronized with those in jni
    public static final int EFFECT_VOLUME = 1;
    public static final int EFFECT_EQ = 2;
    public static final int EFFECT_BALANCE = 3;
    public static final int EFFECT_VIRTUALIZER = 4;
    public static final int EFFECT_BASSBOOST = 5;
    public static final int EFFECT_PRESETREVERB = 6;
    public static final int EFFECT_VISUALIZER = 7;
    public static final int EFFECT_AGC = 8;
    public static final int EFFECT_REPLAYGAIN = 9;
    public static final int EFFECT_SOUNDCHANGE = 10;

    // Effect Param
    /**
     * Number of effect index. eg EQ band
     */
    public static final int PARAM_NUM_INDEXS = 0;
    /**
     * level range.
     */
    public static final int PARAM_LEVEL_RANGE = 1;
    /**
     * level.
     */
    public static final int PARAM_LEVEL = 2;
    /**
     * Band center frequency.
     */
    public static final int PARAM_CENTER_FREQ = 3;
    /**
     * Band frequency range.
     */
    public static final int PARAM_BAND_FREQ_RANGE = 4;
    /**
     * Band for a given frequency.
     * 
     */
    public static final int PARAM_GET_BAND = 5;
    /**
     * Current preset.
     */
    public static final int PARAM_CURRENT_PRESET = 6;
    /**
     * Request number of presets.
     */
    public static final int PARAM_GET_NUM_OF_PRESETS = 7;
    /**
     * Request number of presets.
     */
    public static final int PARAM_PREAMP_LEVEL = 8;

    // ==============for the volume use===============
    // set gain
    // float[0] min(0.0f), float[1] max(1.0f), float[2]ext(1.4f),more than
    // normal max
    public void getVolumeRange(float[] range) {
        int volumeRange[] = new int[3];
        getEffectParamRange(EFFECT_VOLUME, PARAM_LEVEL_RANGE, volumeRange);
        range[0] = volumeRange[0] / 100.0f;
        range[1] = volumeRange[1] / 100.0f;
        range[2] = volumeRange[2] / 100.0f;
    }

    public float getVolume() {
        short nVolume = getEffectParam(EFFECT_VOLUME, PARAM_LEVEL, (short) -1);
        return nVolume / 100.0f;
    }

    /**
     * Sets the volume on this player.
     * 
     * @param volume volume scalar, the value range is 0.0f~1.4f min 0.0f, means mute max 1.0f, means origin level ect
     *            1.4f, means amplify volume set volume from 0.0f~1.0f means change volume from mute to origin level set
     *            volume from 1.0f~1.4f means the volume is amplified, if the origin volume level is low,it is
     *            useful;but is the origin volume level is very high(nearly full scale, the volume may be break )
     * 
     */
    public void setVolume(float volume) {
        short nVolume = (short) (volume * 100);
        setEffectParam(EFFECT_VOLUME, PARAM_LEVEL, (short) -1, nVolume);
    }

    // ================EQ Rated functions==============
    /**
     * Get the number of EQ frequency bands
     * 
     * @return the number of EQ frequency bands current return value is 5.
     */
    public short getEQNumberOfBands() {
        return getEffectParam(EFFECT_EQ, PARAM_NUM_INDEXS, (short) -1);
    }

    /**
     * Get the range of EQ frequency band gain (unit: dB*100)
     * 
     * @param range int[0] min(-1200), int[1] max(1200)
     */
    public void getEQBandLevelRange(int[] range) {
        getEffectParamRange(EFFECT_EQ, PARAM_LEVEL_RANGE, range);
    }

    /**
     * Set the gain in dB*100 of each EQ frequency band
     * 
     * @param band Set the index of the frequency band. The index is between 0 and 'getEQNumberOfBands()-1'
     * @param level The gain value for the band, current range is -1200~1200
     * @throws IllegalStateException
     * @throws IllegalArgumentException
     * @throws UnsupportedOperationException
     */
    public void setEQBandLevel(short band, short level) {
        setEffectParam(EFFECT_EQ, PARAM_LEVEL, band, level);
    }

    /**
     * Get the gain in dB*100 of one EQ frequency band
     * 
     * @param band Set the index of the frequency band. The index is between 0 and 'getEQNumberOfBands()-1'
     * @return The gain value for the band, the range is -1200~1200
     */
    public short getEQBandLevel(short band) {
        return getEffectParam(EFFECT_EQ, PARAM_LEVEL, band);
    }

    public int getEQCenterFreq(short band) {
        short freq = 0;
        freq = getEffectParam(EFFECT_EQ, PARAM_CENTER_FREQ, band);
        return (int) freq;
    }

    // public native void getEQBandFreqRange(short band, int[]range);
    // public native short getEQBand(int frequency);

    /**
     * Set the gain in dB*100 of all EQ frequency bands
     * 
     * @param levles The array size is the band number returned by calling 'getEQNumberOfBands()' the levels range is
     *            -1200~1200
     * @throws IllegalStateException
     * @throws IllegalArgumentException
     * @throws UnsupportedOperationException
     */
    public native void setEQLevels(int[] levels);

    /**
     * Get the gain in dB*100 of all EQ frequency bands
     * 
     * @param levles The array size is the band number returned by calling 'getEQNumberOfBands()' the levels range is
     *            -1200~1200
     */
    public native void getEQLevels(int[] levels);

    /**
     * Get the number of EQ presets
     * 
     * if u want to read chinese please use utf-8 name of presets english-chinese "None" 普�?关闭) "Pop" 流行 "Rock"摇滚
     * "Classical"古典�? * "Jazz"爵士�? * "Folk"民谣 "Voice"人声 "Dance"舞曲 "Treble Boost"高音增强 "Bass Boost"低音增强 "Defalut"默认
     * 
     * @return the number of EQ Presets. we have implemented 11 preset as above lists, So the return number is 11.
     */
    public short getEQNumberOfPresets() {
        return getEffectParam(EFFECT_EQ, PARAM_GET_NUM_OF_PRESETS, (short) -1);
    }

    // default preset index 99
    public static final int DEFAULT_PRESET = 99;
    public static final int EQ_PRESET_NONE = 0;
    public static final int EQ_PRESET_POP = 1;			// 流行
    public static final int EQ_PRESET_ROCK = 2;			// 摇滚
    public static final int EQ_PRESET_CLASSICAL = 3;	// 古典
    public static final int EQ_PRESET_JAZZ = 4;			// 爵士
    public static final int EQ_PRESET_FOLK = 5;			// 民谣
    public static final int EQ_PRESET_VOICE = 6;		// 人声
    public static final int EQ_PRESET_DANCE = 7;		// 舞曲
    public static final int EQ_PRESET_TREBLEBOOST = 8;	// 高音增强
    public static final int EQ_PRESET_BASSBOOST = 9;	// 低音增强

    /**
     * Get the gain in dB*100 of all EQ frequency bands of the preset
     * 
     * @param preset EQ_PRESET_XXX
     * @param levles The array size is the band number returned by calling 'getEQNumberOfBands()' the levels range is
     *            -1200~1200
     */
   public static void getPresetEQLevels(short preset, int[] levels) {
    	native_getPresetEQLevels(preset, levels);
    }
    /**
     * Get the current using EQ preset index.
     * 
     * @return the index of the current using EQ preset, custom if the return is -1 ,from 0 to getEQNumberOfPresets()-2,
     *         DEFAULT_PRESET means default preset.
     */
    public short getEQCurrentPreset() {
        return getEffectParam(EFFECT_EQ, PARAM_CURRENT_PRESET, (short) -1);
    }

    /**
     * Get the EQ preset name.
     * 
     * @param preset The EQ preset index
     * @return the name of the EQ preset.
     * 
     *         Note: if u want to use Chinese name, please ref to the comment of getEQNumberOfPresets().
     */
    public String getEQPresetName(short preset) {
        switch (preset) {
            case 0:
                return "None";
            case 1:
                return "Pop";
            case 2:
                return "Rock";
            case 3:
                return "Classical";
            case 4:
                return "Jazz";
            case 5:
                return "Folk";
            case 6:
                return "Voice";
            case 7:
                return "Dance";
            case 8:
                return "Treble Boost";
            case 9:
                return "Bass Boost";
            case DEFAULT_PRESET:
                return "Default";
            default:
                return "";
        }
    }

    /**
     * Set the EQ preset index to use.
     * 
     * @param preset EQ preset index, EQ_PRESET_NONE...EQ_PRESET_BASSBOOST. DEFAULT_PRESET means default preset.
     */
    public void useEQPreset(short preset) {
        setEffectParam(EFFECT_EQ, PARAM_CURRENT_PRESET, (short) -1, preset);
    }

    /**
     * Get the EQ Preamp value in dB*100. The Preamp works like adding all EQ bands gain with this value.
     * 
     * @return the current using Preamp value, the range is -1200~1200.
     */
    public short getEQPreamp() {
        return getEffectParam(EFFECT_EQ, PARAM_PREAMP_LEVEL, (short) -1);
    }

    /**
     * Set the EQ Preamp value in dB. This function is not implied The Preamp works like adding all EQ bands gain with
     * this value.
     * 
     * @param level Preamp value, the range is -12~12.
     */
    public void setEQPreamp(short level) {
        setEffectParam(EFFECT_EQ, PARAM_PREAMP_LEVEL, (short) -1, level);
    }

    // ==============for the balance use===============
    /**
     * Get the min and max of stereo balance volume.
     * 
     * @param range int[0] min(-100), int[1] max(100).
     */
    public void getBalanceLevelRange(int[] range) {
        getEffectParamRange(EFFECT_BALANCE, PARAM_LEVEL_RANGE, range);
    }

    /**
     * Get the current stereo balance volume.
     * 
     * @return the current stereo balance volume.
     */
    public short getBalanceLevel() {
        return getEffectParam(EFFECT_BALANCE, PARAM_LEVEL, (short) -1);
    }

    /**
     * Set the stereo balance volume.
     * 
     * @param level the stereo balance volume, range is -100~100. -100 means right channel is mute 100 means left
     *            channel is mute 0 means normal volume set level from -100 to 0, the volume of left channel keeps, and
     *            the volume of right channel changes from mute to normal set level from 100 to 0, the volume of right
     *            channel keeps, and the volume of left channel changes from mute to normal
     * @throws IllegalStateException
     * @throws IllegalArgumentException
     * @throws UnsupportedOperationException
     */
    public void setBalanceLevel(short level) {
        setEffectParam(EFFECT_BALANCE, PARAM_LEVEL, (short) -1, level);
    }

    /**
     * Sets the volume on this player. This API is recommended for balancing the output of audio streams within an
     * application.
     * 
     * @param leftVolume left volume scalar, the value range is 0.0f~1.0f
     * @param rightVolume right volume scalar, the value range is 0.0f~1.0f
     */
    public void setVolume(float leftVolume, float rightVolume) {
        float volume = leftVolume;
        short balanceLevel = 0;
        if (rightVolume > leftVolume) {
            volume = rightVolume;
        }
        if (volume > 0) {
            balanceLevel = (short) ((rightVolume - leftVolume) / volume * 100);
        }
        setVolume(volume);
        setBalanceLevel(balanceLevel);
    }

    /**
     * Sets the volume on this player. This API calls setStereoVolume of Audiotrack. This API should be called after
     * Audioplayer is prepared
     * 
     * @param leftVolume left volume scalar, the value range is 0.0f~1.0f
     * @param rightVolume right volume scalar, the value range is 0.0f~1.0f
     */
    public void setVolumeSys(float leftVolume, float rightVolume) {
        mVolumeLeft = leftVolume;
        mVolumeRight = rightVolume;
        setVolumeSys(leftVolume, rightVolume, null);
    }

    // ==============for the virtualizer or surround use===============
    /**
     * 3D Surround effect. A bigger range level has much stronger surround effect.
     * 
     * @param range int[0] min(0), int[1] max(1000).
     * @throws IllegalStateException
     * @throws IllegalArgumentException
     * @throws UnsupportedOperationException
     */
    public void getSurroundLevelRange(int[] range) {
        getEffectParamRange(EFFECT_VIRTUALIZER, PARAM_LEVEL_RANGE, range);
    }

    /**
     * compare to getRoundedStrength of android effect api virtualizer Gets the current strength of the effect.
     * 
     * @return the strength of the effect. The valid range for strength is [0, 1000], where 0 per mille designates the
     *         mildest effect and 1000 per mille the strongest
     */
    public short getSurroundLevel() {
        return getEffectParam(EFFECT_VIRTUALIZER, PARAM_LEVEL, (short) -1);
    }

    /**
     * compare to setStrength of android effect api virtualizer Sets the strength of the virtualizer Surround effect. * @param
     * strength strength of the effect. The valid range for strength strength is [0, 1000], where 0 per mille designates
     * the mildest effect and 1000 per mille designates the strongest.
     * 
     * @throws IllegalStateException
     * @throws IllegalArgumentException
     * @throws UnsupportedOperationException
     */
    public void setSurroundLevel(short level) {
        setEffectParam(EFFECT_VIRTUALIZER, PARAM_LEVEL, (short) -1, level);
    }

    // ==============for the bass boost use===============
    /**
     * bass boost effect. A bigger range level has much stronger surround effect.
     * 
     * @param range int[0] min(0), int[1] max(1000).
     */
    public void getBassLevelRange(int[] range) {
        getEffectParamRange(EFFECT_BASSBOOST, PARAM_LEVEL_RANGE, range);
    }

    /**
     * compare to getRoundedStrength of android effect api bassboost Gets the current strength level of the effect.
     * 
     * @return the strength of the effect. The valid range for strength is [0, 1000], where 0 per mille designates the
     *         mildest effect and 1000 per mille the strongest
     */
    public short getBassLevel() {
        return getEffectParam(EFFECT_BASSBOOST, PARAM_LEVEL, (short) -1);
    }

    /**
     * compare to setStrength of android effect api bassboost Sets the strength level of the bass boost effect.
     * 
     * @param strength strength of the effect. The valid range for strength strength is [0, 1000], where 0 per mille
     *            designates the mildest effect and 1000 per mille designates the strongest.
     */
    public void setBassLevel(short level) {
        setEffectParam(EFFECT_BASSBOOST, PARAM_LEVEL, (short) -1, level);
    }

    // ==============for the visual use===============
    /**
     * get frequency data range by analyzing the song.
     * 
     * @param range short[0] min(0), short[1] max(100).
     */
    public native void getFreqValueRange(int[] range);

    /**
     * get frequency data by analyzing the song.
     * 
     * @param input :datacount the count of frequency data u want to get.
     * @param output :data stores the frequency data u need, memory should be managed by user.
     */
    // 获得指定数目的频谱数据，用户需要的数据量即实际需要绘制的数目，data为大小为datacount的数量

    public void getFreqData(int datacount, int[] data) {

        AudioTrack audioTrack = mAudioTrack;
        try {
            if (audioTrack != null) {
                /*
                 * int nbitpersample = 2; if (audioTrack.getAudioFormat() == AudioFormat.ENCODING_PCM_8BIT) {
                 * nbitpersample = 1; }
                 * 
                 * _getFreqData( audioTrack.getPlaybackHeadPosition() audioTrack.getChannelCount() * nbitpersample,
                 * datacount, data);
                 */
                int lAudioTrackByte = (int) mAudioTrackSum;
                if (lAudioTrackByte >= mBufferSize)// delay between write and
                                                   // sound
                    lAudioTrackByte -= mBufferSize;

                _getFreqData(lAudioTrackByte, datacount, data);
            } else {
                for (int i = 0; i < datacount; ++i) {
                    data[i] = 0;
                }
            }
        } catch (Exception ex) {
            logHelper.errorShow(ex.toString());
        }
    }

    public void enableFadeInFadeOut(boolean bEnable) {
        logHelper.infoShow("enableFadeInFadeOut " + bEnable);
        mIsEnableFadeInFadeOut = bEnable;
    }

    public static native String getVersion();

    public static native void setPlayerFlag(int flag);

    // enable or disable effect, this version(3.5) reference to EFFECT_EQ,
    // EFFECT_REPLAYGAIN and EFFECT_VIRTUALIZER
    public native void enableEffect(int nEffectType, boolean bEnable);

    /**
     * 关闭所有音效或恢复所有音效。(不包含音量平衡音效，音量平衡单独控制)
     * 主动关闭调音台，设备音效，推荐音效时传入true
     * 恢复关闭调音台，设备音效，推荐音效时传入false。
     * @param  bBypass true:关闭所有音效，关闭后设置音效参数无效 false:打开所有音效，各参数保持
     */
    public native void bypassResumeAllEffects(boolean bBypassOrResume);

    /**
     * 设置组合音效，用于推荐音效，汽车音效，明星音效等，传入一个音效参数文件。如果传入null表示重置组合音效。
     * 注意 组合音效切换到其他非组合音效时需要调用setAudioEffectCombine(null)重置组合音效
     * @param  path 音效参数文件，如果传入null表示重置组合音效。
     * @throws IllegalStateException
     * @throws IllegalArgumentException
     * @throws RuntimeException
     */
    public void setAudioEffectCombine(String aebFilePath)  throws IllegalStateException, IllegalArgumentException, RuntimeException {
        native_setAudioEffectCombine(aebFilePath);
    }
    /**
     * Gets the current strength level of the replaygain effect. in dB*100
     * 
     * @return current replaygain [-1200,1200]
     */
    public short getReplayGain() {
        return getEffectParam(EFFECT_REPLAYGAIN, PARAM_LEVEL, (short) -1);
    }

    /**
     * Sets the strength level of the replaygain effect.in dB*100
     * 
     * @param level of the effect. The valid range for strength strength is [-1200, 1200],
     */
    public void setReplayGain(short level) {
        setEffectParam(EFFECT_REPLAYGAIN, PARAM_LEVEL, (short) -1, level);
    }

    /**
     * Get the range of replaygain (unit: dB*100)
     * 
     * @param range int[0] min(-1200), int[1] max(1200)
     */
    public void getReplayGainRange(int[] range) {
        getEffectParamRange(EFFECT_REPLAYGAIN, PARAM_LEVEL_RANGE, range);
    }

    /**
     * Get extra trace Info when native crash
     * 
     * @return Crash Info String (audio url, machine, userid, version...)
     */
    public static String getCrashTraceInfo() {
        return native_getCrashTraceInfo();
    }

    public static void onNativeCrashed(String strExceptionDesc) {
        String strExceptionOut =
                "catch a audiocore_native_crashed named " + strExceptionDesc
                        + " from jni and print current java stack info...";

        logHelper.errorShow(strExceptionOut);
        throw new IllegalStateException("audiocore_native_crashed");
    }

    // ==============for reverb environment effect===============
    /**
     * Get the number of reverb effect presets
     * 
     * preset names 0 无 1 演唱会 2 卧室 3 石屋 4 剧院 5 洞穴 6 空巷 7 城市 8 停车场 9 客厅 10 礼堂 11 大殿
     * 
     * @return the number of environment effect Presets. we have implemented 12 preset as above lists, So the return
     *         number is 12.
     */
    public short getReverbNumberOfPresets() {
        return getEffectParam(EFFECT_PRESETREVERB, PARAM_GET_NUM_OF_PRESETS, (short) -1);
    }

    /**
     * Get preset of reverb.
     * 
     * @return reverb preset index, from 0 to getReverbNumberOfPresets()-1.
     */
    public short getReverbCurrentPreset() {
        return getEffectParam(EFFECT_PRESETREVERB, PARAM_CURRENT_PRESET, (short) -1);
    }

    /**
     * Enables a preset on the reverb.
     * 
     * @param reverb preset index from 0 to getReverbNumberOfPresets()-1, refrence to "preset names".
     */
    public void useReverbPreset(short preset) {
        setEffectParam(EFFECT_PRESETREVERB, PARAM_CURRENT_PRESET, (short) -1, preset);
    }

    /**
     * set playback speed
     * @param speed 0.5f - 2.0f
     */
    public void setPlaybackSpeed(float speed) {
    	mPlaybackSpeed = speed;
    	_setSpeed();
    }
    
    /**
     * get current playback speed
     * @return current speed
     */
    public float getPlaybackSpeed() {
    	return mPlaybackSpeed;
    }
    
    /**
     * Gets audio duration by audio file path .
     * 
     * @param filePath the Context to use when resolving the Uri
     * @throws IOException if it is called in an invalid state
     * @throws IllegalStateException if it is called in an invalid state
     * @throws IllegalArgumentException if it is called in an invalid state
     */
    public static int getAudioDuration(String filePath) throws IOException, IllegalStateException,
            IllegalArgumentException {
        logHelper.infoShow("getAudioDuration start ");
        return native_getAudioDuration(filePath);
    }

    
    public static boolean getaudiotags(ACAudioTags tag)  throws IOException, IllegalStateException,
    		IllegalArgumentException {
        logHelper.infoShow("getAudioTags start ");
        ACAudioTags newTags = native_getAudioTags(tag.filePath, tag.imagePath, tag.lrcPath);
        if(null == newTags) {
        	logHelper.infoShow("getAudioTags failed ");
        	return false;
        }
        
        tag.copyInfo(newTags);
        return true;
    }
    
    /*
     * inner java
     */

    protected void _setSpeed() {
    	if(null == mAudioTrack) {
    		return;
    	}
    	
		// play back rate > 6.0
//		if (android.os.Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
//			android.media.PlaybackParams playbackParams = new android.media.PlaybackParams();
//			playbackParams.setSpeed(mPlaybackSpeed);
//			mAudioTrack.setPlaybackParams(playbackParams);
//		} else {
			short level = (short) (100 * mPlaybackSpeed);
			setEffectParam(EFFECT_SOUNDCHANGE, PARAM_LEVEL, (short) -1, level);
//		}
    }
    
    protected boolean advertisementSendSingal(AdvertiseMentInfoReport param) {
        return native_advertisementSendSingal(param);
    }

    // private static String getSystemInfo() {
    // return android.os.Build.MODEL + "," + android.os.Build.VERSION.SDK
    // + "," + android.os.Build.VERSION.RELEASE;
    // }

    private static BroadcastReceiver connectionReceiver = new BroadcastReceiver() {

        private int getNetworkClass(int networkType) {

            if (Build.VERSION.SDK_INT >= 14) {// 13 is ok
                switch (networkType) {
                    case TelephonyManager.NETWORK_TYPE_GPRS:
                    case TelephonyManager.NETWORK_TYPE_EDGE:
                    case TelephonyManager.NETWORK_TYPE_CDMA:
                    case TelephonyManager.NETWORK_TYPE_1xRTT:
                    case TelephonyManager.NETWORK_TYPE_IDEN:
                        return NETWORK_TYPE_2G;
                    case TelephonyManager.NETWORK_TYPE_UMTS:
                    case TelephonyManager.NETWORK_TYPE_EVDO_0:
                    case TelephonyManager.NETWORK_TYPE_EVDO_A:
                    case TelephonyManager.NETWORK_TYPE_HSDPA:
                    case TelephonyManager.NETWORK_TYPE_HSUPA:
                    case TelephonyManager.NETWORK_TYPE_HSPA:
                    case TelephonyManager.NETWORK_TYPE_EVDO_B:
                    case TelephonyManager.NETWORK_TYPE_EHRPD:
                    case TelephonyManager.NETWORK_TYPE_HSPAP:
                        return NETWORK_TYPE_3G;
                    case TelephonyManager.NETWORK_TYPE_LTE:
                        return NETWORK_TYPE_4G;
                    default:
                        return NETWORK_TYPE_UNKNOWN;
                }
            } else {
                if (networkType == TelephonyManager.NETWORK_TYPE_GPRS
                        || networkType == TelephonyManager.NETWORK_TYPE_CDMA
                        || networkType == TelephonyManager.NETWORK_TYPE_EDGE) {
                    return NETWORK_TYPE_2G;
                } else {
                    return NETWORK_TYPE_3G;
                }
            }

        }

        // private int Is2GOr3G(int netType) {
        // if (netType == TelephonyManager.NETWORK_TYPE_GPRS
        // || netType == TelephonyManager.NETWORK_TYPE_CDMA
        // || netType == TelephonyManager.NETWORK_TYPE_EDGE) {
        // return NETWORK_TYPE_2G;
        // } else {
        // return NETWORK_TYPE_3G;
        // }
        // }

        @Override
        public void onReceive(Context context, Intent intent) {
            // TODO Auto-generated method stub
            if (longLifeContext == null) {
                return;
            }

            ConnectivityManager connectMgr =
                    (ConnectivityManager) longLifeContext.getSystemService(Context.CONNECTIVITY_SERVICE);
            if (connectMgr == null) {
                return;
            }

            NetworkInfo mobNetInfo = connectMgr.getNetworkInfo(ConnectivityManager.TYPE_MOBILE);
            if (mobNetInfo == null) {
                return;
            }

            NetworkInfo wifiNetInfo = connectMgr.getNetworkInfo(ConnectivityManager.TYPE_WIFI);
            if (wifiNetInfo == null) {
                return;
            }

            if (!mobNetInfo.isConnected() && !wifiNetInfo.isConnected()) {
                mNetState = NETWORK_TYPE_NONE;
                logHelper.debugShow("unconnected");
                // unconnect network
            } else {
                // connect network
                NetworkInfo info = connectMgr.getActiveNetworkInfo();
                if (info != null) {
                    switch (info.getType()) {
                        case ConnectivityManager.TYPE_WIFI:// wifi
                            mNetState = NETWORK_TYPE_WIFI;
                            logHelper.debugShow("wifi");
                            break;
                        case ConnectivityManager.TYPE_MOBILE: // mobile
                            mNetState = getNetworkClass(info.getSubtype());
                            logHelper.debugShow("mobile");
                            break;
                        default:
                            mNetState = NETWORK_TYPE_NONE;
                            break;
                    }
                }

            }

            setUpDownloadTypeByNetType();
            setNetType(mNetState);
        }
    };

    protected boolean advertisementSetInfos(AdvertiseMentInfos infos) {
        return native_advertisementSetInfos(infos);
    }

    protected boolean advertisementSetAdInfos(AdvertiseMentInfos infos) {
        return native_advertisementSetAdInfos(infos);
    }

    private int checkAdInterrupt() {
        int adposition = 0;
        if ((mPolicy != null) && (mPolicy.pAdPlayingValid) && (mBytePerSec > 0)) {
            adposition = (int) ((mPolicy.pSingalAdSumSize * 1000) / mBytePerSec);
            logHelper.infoShow(" policy sendAdInterrupt adposition " + adposition);

            // report audiocore;
            formatAdSendSignal(AD_INTERRUPT);

            if ((mOnAdvertiseMentPlayStatusListener != null) && (mAdCurAdIndex >= 0)
                    && (mAdCurAdIndex < mAdInfos.adNums) && (mAdInfos.adId != null)) {
                mOnAdvertiseMentPlayStatusListener.onAdvertiseMentPlayStatus(AudioPlayer.this, AD_INTERRUPT,
                        mAdInfos.adId[mAdCurAdIndex], adposition > 0 ? adposition : 0);
            }

            resetAdStatus();
        }
        return 0;
    }

    private int InitAdParm() {
        synchronized (lockAd) {
            int sampleRateInHz = this.getRenderSampleRate();
            int channelConfig = this.getRenderChannels();
            int audioFormat = this.getRenderBitsPerSample();
            int bytepersec = sampleRateInHz * channelConfig * audioFormat / 8;

            if ((mPolicy != null) && (bytepersec > 0)) {
                mPolicy.pSingalSongValidSize = mPolicy.pSingalSongValidTime * bytepersec;
                mPolicy.pSingalSongCalcValid = true;
                mPolicy.pSingalSongSumSize = 0;
                mPolicy.pSingalSongSumLastSize = 0;
                // 统一在reset中作了广告播放状态设置
                // mPolicy.pAdPlayingValid = false;
                mPolicy.pSingalAdSumSize = 0;
                mPolicy.pMultiAdSumSize = 0;

                if (mAdFirstInsertValid == false) {
                    mPolicy.pFirstInsertSongFinished = true;
                    mAdTimerRelativeMaxTime = mPolicy.pMaxTotalTime;
                    logHelper.infoShow("policy - checktotaltime maxtime curtime(" + mPolicy.pSingalSongValidCount + ","
                            + mAdTimerRelativeMaxTime + "," + b2ms(mAdTimerRelativeSize) / 1000 + ")");

                }
                if ((mAdInfos != null) && (mAdInfos.adNums > mAdInfos.anchorNums)) {
                    for (int i = 0; i < mAdInfos.adNums; i++) {
                        mAdMultiAdDurationSize[i] = (long) ((long) mAdInfos.adDuration[i] * (long) bytepersec) / 1000;
                        logHelper.infoShow("policy adindex duration(" + i + "," + mAdInfos.adDuration[i] + ","
                                + mAdMultiAdDurationSize[i] + ")");
                    }
                }
                logHelper.infoShow(" policy-init first times Maxsongs volidtimes flag Maxflag("
                        + mPolicy.pFirstInsertSongNum + "," + mPolicy.pSingalSongValidTime + ","
                        + mPolicy.pSingalSongMaxCount + "," + mPolicy.pMaxTotalTime + ","
                        + mPolicy.pFirstInsertSongFinished + "," + mPolicy.pSingalSongMaxCountCalcValid + ")");
            }

            mAdTimerRelativeMaxSize = (long) bytepersec * (long) mAdTimerRelativeMaxTime;

            // 插入频率较高时预下载时机调整为最大时长1/4
            long calcPreDownloadTime =
                    ((mAdPreDownloadTime > mAdTimerRelativeMaxTime / 4) && (mAdTimerRelativeMaxTime != 0)) ? mAdTimerRelativeMaxTime / 4
                            : mAdPreDownloadTime;
            mAdPreDownloadSize = (long) bytepersec * calcPreDownloadTime;

            // 插入频率较高时插入时机调整为不提前
            long calcInsertTime =
                    ((mAdPreInsertTime > mAdTimerRelativeMaxTime) && (mAdTimerRelativeMaxTime != 0)) ? 0
                            : mAdPreInsertTime;
            mAdPreInsertSize = (long) bytepersec * calcInsertTime;

            // 插入频率较高时预下载时机调整为最大时长一半
            long calcPrePreDownloadTime =
                    ((mAdPrePreDownloadTime > mAdTimerRelativeMaxTime / 2) && (mAdTimerRelativeMaxTime != 0)) ? mAdTimerRelativeMaxTime / 2
                            : mAdPrePreDownloadTime;
            mAdPrePreDownloadSize = (long) bytepersec * calcPrePreDownloadTime;

            logHelper.infoShow("policy Adinit pre download insert (" + mAdPreDownloadSize + "," + mAdPreDownloadTime
                    + "," + mAdPreInsertSize + "," + mAdPreInsertTime + "," + mAdPrePreDownloadSize + ","
                    + mAdPrePreDownloadTime + ")");
            logHelper.infoShow("Adinit byteps maxtime(" + bytepersec + "," + mAdTimerRelativeMaxTime + ")");

            if ((mAdInfos != null) && (mAdInfos.anchorNums > 0)) {
                for (int i = 0; i < mAdInfos.anchorNums; i++) {
                    mAdCurAnchorSize = (long) ((long) mAdInfos.anchorTime[i] * (long) bytepersec) / 1000;
                    logHelper.infoShow("adinit anchorpos time (" + i + "," + mAdCurAnchorSize + ","
                            + mAdInfos.anchorTime[i] + ")");
                }
            }
        }

        return 0;
    }

    private int resetAdStatus() {
        synchronized (lockAd) {
            // lebo
            mAdCurAdIndex = 0;
            mAdTimerRelativeSize = 0;
            mAdSingalPreDownloadValid = true;
            mAdSingalDownloadValid = true;
            mAdSingalInsertValid = true;
            mAdCurAnchorValid = false;
            mAdWillPlayValid = false;
            mAdReportMessageValid = true;

            if (mPolicy != null) {
                mPolicy.pSingalSongCalcValid = true;
                mPolicy.pSingalSongMaxCountCalcValid = true;

                mPolicy.pSingalSongSumSize = 0;
                mPolicy.pSingalSongSumLastSize = 0;

                mPolicy.pSingalSongValidCount = 0;
                mPolicy.pSingalAdSumSize = 0;
                mPolicy.pMultiAdSumSize = 0;

                mPolicy.pAdPlayingValid = false;
            }

            mAdInsertConditionSingalSongMaxCountIsValid = false;

            logHelper.infoShow(" policy-finish_statusreset (" + b2ms(mAudioTrackSumLast) + "," + b2ms(mAdTimerSize)
                    + "," + b2ms(mAudioTrackSum) + "," + b2ms(mAdCurAdDurationSize) + ","
                    + b2ms(mAdAbsoluteDurationSize) + "," + mAdInsertConditionSingalSongMaxCountIsValid + ")");
        }

        return 0;
    }

    private long multiAdCalcDuration(int anchorindex, long curanchorsize) {
        if ((mPolicy == null) || (mAdInfos == null)) {
            logHelper.infoShow("multiAdCalcDuration Failed.");
            return -1;
        }
        long adtotaldurationsize = 0;
        for (int i = 0; i < (mAdInfos.anchorNumsOfAd[anchorindex]); i++) {
            adtotaldurationsize += mAdMultiAdDurationSize[i];
        }

        for (int i = 0; i < mAdInfos.anchorNumsOfAd[mAdCurAnchorIndex]; i++) {
            mAdMultiAbsoluteDurationSize[i] = curanchorsize + mAdMultiAdDurationSize[i];
            curanchorsize = mAdMultiAbsoluteDurationSize[i];
            // logHelper.infoShow(" policy-calcmulti duration  ("
            // + i+ ","
            // + mAdMultiAdDurationSize[i] + ","
            // + mAdMultiAbsoluteDurationSize[i] + ","
            // + mAdCurAdDurationSize + ")");
        }
        return adtotaldurationsize;
    }

    private int multiAdFindBestAdIndex(int anchorindex) {
        int nums = 0;
        int adindex = 0;
        if ((mPolicy == null) || (mAdInfos == null)) {
            logHelper.infoShow("multiAdFindBestAdIndex Failed.");
            return -1;
        }
        nums = mAdInfos.anchorNums;
        for (int i = 0; i < nums; i++) {
            if (i >= anchorindex) {
                return adindex;
            }
            adindex += mAdInfos.anchorNumsOfAd[i];
        }
        return -1;
    }

    protected long b2ms(long nbyte) {
        if (mBytePerSec == 0) {
            return 0;
        }
        return (long) ((nbyte * 1000) / mBytePerSec);
    }

    private class RenderHandlerThread extends HandlerThread implements Callback {
        private byte[] mData;

        // private boolean mIsSetPriority;

        public RenderHandlerThread(String name) {
            super(name);
            // setPriority(MAX_PRIORITY); // set to normal - 1
            mData = new byte[4096];
            // mIsSetPriority = false;
        }

        @Override
        public boolean handleMessage(Message msg) {
            // android.os.Process.THREAD_PRIORITY_URGENT_AUDIO is too high,
            // some synchronized may be blocked,eg FADE_MODE_STOP
            // if(!mIsSetPriority) {
            // int nTid = android.os.Process.myTid();
            // int nOriPriorty = android.os.Process.getThreadPriority(nTid);
            // // 设置线程优先级，保证资源分配。
            // android.os.Process
            // .setThreadPriority(android.os.Process.THREAD_PRIORITY_URGENT_AUDIO);
            //
            // int nCurPriorty = android.os.Process.getThreadPriority(nTid);
            //
            // logHelper.infoShow("RenderHandlerThreadPriority nTid "+ nTid +
            // "nOriPriorty " + nOriPriorty + "nCurPriorty" + nCurPriorty);
            //
            // mIsSetPriority = true;
            // }
            if (msg.what == RENDER_RUN) {
                int sizeInBytes = getRenderBuffer(mBufferSize < mData.length ? mBufferSize : mData.length, mData);
                // xiaomi 3
                // 和其他手机调用audiotrack方式不太一样，其他手机均匀调用audiotrack的write，小米3连续调用N（N大概为7）次后再休息200毫秒，
                // 由于底层buffer有限，被瞬时耗尽，导致getRenderBuffer每轮的最后几次可能取不到数据,会有buffer不足日志打印，但是不会造成卡断
                // logHelper.infoShow("getRenderBuffer " + sizeInBytes);
                if (sizeInBytes > 0) {
                    // logHelper.infoShow("AudioTrack write begin.");
                    // ZTE blocks when
                    // getPlaybackHeadPosition

                    if (mOnGetRendderBufferListener != null) {
                        mOnGetRendderBufferListener.onGetRenderBuffer(sizeInBytes, mData, AudioPlayer.this);
                    }

                    if (sizeInBytes <= 10240) {
                        synchronized (AudioPlayer.this) {
                            if (mAudioTrack != null) {// mingliang

                                // FADE_MODE_STOP will set mAudioTrack to null,
                                // protect although already locked
                                try {
                                    mAudioTrack.write(mData, 0, sizeInBytes);
                                } catch (Exception e) {
                                    e.printStackTrace();
                                }
                            } else {
                                logHelper.debugShow("mAudioTrack = null, sizeInBytes=%d" + sizeInBytes);
                            }
                        }

                        // 广告场景和非广告场景
                        if ((mPolicy != null) || (mAdInfos != null)) {
                            calcAdWhenPlaying(sizeInBytes);
                        } else {
                            // < 6.0 use soundtouch, position need recalc
//                    		if ((android.os.Build.VERSION.SDK_INT < Build.VERSION_CODES.M) && Math.abs(mPlaybackSpeed - 1.0f) > 0.000001) {
                    			sizeInBytes *= mPlaybackSpeed;
//                    		}

                            mAudioTrackSum += sizeInBytes;
                        }
                    }

                    // logHelper.infoShow("AudioTrack write end.");

                    Handler Renderhandler = mRenderhandler;
                    if (Renderhandler != null) {
                        Renderhandler.sendEmptyMessage(RENDER_RUN);
                    }

                } else {
                    // Wait 1ms.
                    Handler Renderhandler = mRenderhandler;
                    logHelper.infoShow("render sleep 5ms ");
                    if (Renderhandler != null) {
                        Renderhandler.sendEmptyMessageAtTime(RENDER_RUN, SystemClock.uptimeMillis() + 5);
                        // Renderhandler.sendEmptyMessageDelayed(RENDER_RUN,5);
                    }
                    // try {
                    // Thread.sleep(10);
                    // } catch (InterruptedException e) {
                    // e.printStackTrace();
                    // }
                    // mRenderhandler.sendEmptyMessage(RENDER_RUN);
                }
            } else if (msg.what == RENDER_START) {
                logHelper.infoShow("render RENDER_START start");
                synchronized (AudioPlayer.this) {
                    Handler Renderhandler = mRenderhandler;
                    if (Renderhandler != null) {
                        Renderhandler.sendEmptyMessage(RENDER_RUN);
                    }
                }
                logHelper.infoShow("render RENDER_START end");
            } else if (msg.what == RENDER_STOP) {
                logHelper.infoShow("render RENDER_STOP start");
                synchronized (AudioPlayer.this) {
                    Handler Renderhandler = mRenderhandler;
                    if (Renderhandler != null) {
                        Renderhandler.removeMessages(RENDER_RUN);
                    }
                }
                logHelper.infoShow("render RENDER_STOP end");
            }
            return true;
        }

        boolean calcAdWhenPlaying(int sizeInBytes) {
            synchronized (lockAd) {
                mAdTimerSize += sizeInBytes;

                // calc absolute duration ,and ad display
                if (mAdCurAnchorValid && (mAdInfos != null) && (mAdInfos.anchorNums > 0) && (mAdInfos.adNums > 0)) {
                    mAdCurAnchorSize =
                            (long) ((long) mAdInfos.anchorTime[mAdCurAnchorIndex] * (long) mBytePerSec) / 1000;
                    // multi ad calc every ad duration size;
                    if (mAdInfos != null) {
                        if (mAdInfos.anchorNumsOfAd[mAdCurAnchorIndex] > 1) {
                            // mAdCurAdIndex =
                            // multiAdFindBestAdIndex(mAdCurAnchorIndex);
                            mAdCurAdDurationSize = multiAdCalcDuration(mAdCurAnchorIndex, mAdCurAnchorSize);
                        } else {
                            mAdCurAdDurationSize =
                                    (long) ((long) mAdInfos.adDuration[mAdCurAdIndex] * (long) mBytePerSec) / 1000;
                        }
                    }
                    mAdAbsoluteDurationSize = mAdCurAnchorSize + mAdCurAdDurationSize;
                    // logHelper.infoShow(" policy-calcmulti durationsize adindex addusize ("
                    // + mAdTimerSize + ","
                    // + mAdCurAnchorSize + ","
                    // + mAdAbsoluteDurationSize + ","
                    // + mAdCurAdIndex + ","
                    // + mAdCurAnchorValid + ","
                    // + mAdReportMessageValid + ","
                    // + mAdCurAdDurationSize + ")");
                }
                // play ad
                if (mAdCurAnchorValid && (mAdTimerSize >= mAdCurAnchorSize)
                        && (mAdTimerSize <= mAdAbsoluteDurationSize)) {
                    if (mAdReportMessageValid) {
                        // adplaying
                        advertiseMentSendMessage(AD_PLAYING, mAdCurAdIndex);
                        mAdReportMessageValid = false;
                        // if ad played ,not run (2,3,0)- 2;
                        mAdhavePlayedValid = true;
                    }

                    // multi ad playing,send finished ;
                    if ((mAdInfos != null) && (mAdInfos.anchorNumsOfAd[mAdCurAnchorIndex] > 1)
                            && ((mAdCurAdIndex + 1) < mAdInfos.anchorNumsOfAd[mAdCurAnchorIndex])) {
                        if (mAdTimerSize > mAdMultiAbsoluteDurationSize[mAdCurAdIndex]) {
                            advertiseMentSendMessage(AD_FINISHED, mAdCurAdIndex);
                            logHelper.infoShow(" policy-multiSingalFinished  (" + mAdAbsoluteDurationSize + ","
                                    + mAdTimerSize + "," + mAdMultiAdDurationSize[mAdCurAdIndex] + "," + mAdCurAdIndex
                                    + ")");
                            mAdCurAdIndex = mAdCurAdIndex + 1;
                            mPolicy.pSingalAdSumSize = 0;
                            advertiseMentSendMessage(AD_PLAYING, mAdCurAdIndex);// next
                                                                                // playing;
                            logHelper.infoShow(" policy-multiSingalNext  (" + mAdAbsoluteDurationSize + ","
                                    + mAdTimerSize + "," + mAdMultiAdDurationSize[mAdCurAdIndex] + "," + mAdCurAdIndex
                                    + ")");
                        }
                    }
                    mAudioTrackSum = mAudioTrackSumLast;
                    mAdTimerRelativeSize = mAdTimerRelativeLastSize;
                    if (mPolicy != null) {
                        // calc ad play position regardless of
                        // mPolicy.bNeedCalcAdProgress
                        mPolicy.pSingalSongSumSize = mPolicy.pSingalSongSumLastSize;
                        mPolicy.pSingalAdSumSize += sizeInBytes;
                        mPolicy.pMultiAdSumSize += sizeInBytes;

                        mPolicy.pAdPlayingValid = true;
                    }

                } else {
                    // check if ad will be finished ;
                    if (mAdCurAnchorValid && (mAdTimerSize > mAdAbsoluteDurationSize) && (mPolicy != null)
                            && (mPolicy.pAdPlayingValid)) {
                        mAdTimerSize = mAdTimerSize - mAdCurAdDurationSize; // get
                                                                            // rid
                                                                            // of
                                                                            // adduration;
                        advertiseMentSendMessage(AD_FINISHED, mAdCurAdIndex);
                        advertiseMentSendMessage(AD_ALLFINISHED, mAdCurAdIndex);
                        resetAdStatus();
                        // only ad lebo
                        if ((mAdInfos != null) && (mAdInfos.anchorNumsOfAd[mAdCurAnchorIndex] == 1)) {
                            if (mAdCurAdIndex + 1 < mAdInfos.adNums) {
                                mAdCurAdIndex = mAdCurAdIndex + 1;// upgradeadindex;
                            } else {
                                mAdCurAdIndex = 0;
                            }
                        }
                        logHelper.infoShow("advertiseupgrade nextAdindex=" + mAdCurAdIndex);
                    }

                    // play song;
                    mAudioTrackSum += sizeInBytes;
                    mAudioTrackSumLast = mAudioTrackSum;
                    mAdTimerRelativeSize += sizeInBytes;
                    mAdTimerRelativeLastSize = mAdTimerRelativeSize;
                    if (mPolicy != null) {
                        // calc ad play position regardless of
                        // mPolicy.bNeedCalcAdProgress
                        mPolicy.pSingalSongSumSize += sizeInBytes;
                        mPolicy.pSingalSongSumLastSize = mPolicy.pSingalSongSumSize;

                        mPolicy.pAdPlayingValid = false;
                    }
                }

                // insert first condition
                if (mPolicy != null && (mPolicy.pSingalSongCalcValid)) {
                    // calc if singal song valid;
                    if (mPolicy.pSingalSongSumSize > mPolicy.pSingalSongValidSize) {
                        mPolicy.pSingalSongValidCount++;
                        mPolicy.pSingalSongCalcValid = false;
                        logHelper.infoShow(" policy-valid count add validtime (" + mPolicy.pSingalSongValidCount + ","
                                + b2ms(mPolicy.pSingalSongSumSize) + "," + b2ms(mPolicy.pSingalSongValidSize) + ")");
                        logHelper.infoShow(" policy-valid first flag  nums (" + mAdFirstInsertValid + ","
                                + mPolicy.pFirstInsertSongNum + ")");
                    }
                    // first insert finished , reset singsong validcount;send
                    // perdownload
                    if ((mAdFirstInsertValid == true) && (mPolicy.pFirstInsertSongNum >= 2)
                            && (mPolicy.pSingalSongValidCount >= (mPolicy.pFirstInsertSongNum - 1))) {
                        mAdFirstInsertValid = false;
                        mPolicy.pSingalSongValidCount = 0;
                        // first arrive app / pause time > 30 mins
                        mAdInsertConditionSingalSongMaxCountIsValid = true;
                        advertiseMentSendMessage(AD_PREDOWNLOAD, mAdCurAdIndex);
                        logHelper.infoShow(" policy-first insertfinished  validcount firstsongnums("
                                + mPolicy.pSingalSongValidCount + "," + mPolicy.pFirstInsertSongNum + ")");
                    }
                    // after first insert finished ,if the count of valicsong >
                    // max , singalsong max count is valid;
                    // or if each audio need insert ad, predownload when
                    // (mAdTimerRelativeSize >= (mAdTimerRelativeMaxSize -
                    // mAdPrePreDownloadSize))
                    if ((mPolicy.pFirstInsertSongFinished) && (mPolicy.pSingalSongMaxCountCalcValid)
                            && ((mPolicy.pSingalSongValidCount > (mPolicy.pSingalSongMaxCount - 1)))) {
                        mPolicy.pSingalSongMaxCountCalcValid = false;
                        mAdInsertConditionSingalSongMaxCountIsValid = true;
                        logHelper.infoShow(" policy-Max count Maxcount curtime (" + mPolicy.pSingalSongValidCount + ","
                                + mPolicy.pSingalSongMaxCount + "," + b2ms(mAdTimerRelativeSize) / 1000 + ")");
                    }
                }

                // calc predownload/download/insert when mPolicy != null
                if (mPolicy != null) {
                    // second condition;
                    if (mAdSingalPreDownloadValid && mAdInsertConditionSingalSongMaxCountIsValid
                            && (mAdTimerRelativeSize >= (mAdTimerRelativeMaxSize - mAdPrePreDownloadSize))
                            && (mPolicy.pFirstInsertSongFinished)) { // if not
                                                                     // add
                                                                     // mPolicy.pFirstInsertSongFinished,
                                                                     // AD_PREDOWNLOAD
                                                                     // will
                                                                     // receive
                                                                     // twice
                                                                     // if
                                                                     // mAdFirstInsertValid
                                                                     // is
                                                                     // true
                        mAdSingalPreDownloadValid = false;
                        advertiseMentSendMessage(AD_PREDOWNLOAD, mAdCurAdIndex);
                    }

                    // send download;
                    if (mAdSingalDownloadValid
                            && (mAdTimerRelativeSize >= (mAdTimerRelativeMaxSize - mAdPreDownloadSize))
                            && mAdInsertConditionSingalSongMaxCountIsValid) {
                        advertiseMentSendMessage(AD_DOWNLOAD, mAdCurAdIndex);
                        mAdSingalDownloadValid = false;
                        logHelper.infoShow("policy SingalDownload mAdTimerAbsolute " + b2ms(mAdTimerRelativeSize)
                                + b2ms(mAdTimerRelativeMaxSize - mAdPreDownloadSize));
                    }
                    // send insert;
                    if (mAdSingalInsertValid && (mAdTimerRelativeSize >= (mAdTimerRelativeMaxSize - mAdPreInsertSize))
                            && mAdInsertConditionSingalSongMaxCountIsValid) {
                        // if first insert not finished ,can not send insert;
                        if ((mPolicy == null) || ((mPolicy != null) && (0 == mPolicy.pFirstInsertSongNum))
                                || ((mPolicy != null) && (mPolicy.pFirstInsertSongFinished))) {
                            logHelper.infoShow("policy SingalInsert mAdTimerAbsolute " + b2ms(mAdTimerRelativeSize)
                                    + b2ms(mAdTimerRelativeMaxSize - mAdPreInsertSize));
                            advertiseMentSendMessage(AD_INSERT, mAdCurAdIndex);
                        }
                        mAdSingalInsertValid = false;
                        mAdWillPlayValid = true;
                        logHelper.infoShow("policy mAdWillPlayValid = true ");
                    }
                }
            }

            return true;
        }
    }

    @Override
    protected void finalize() {
        native_finalize();
    }

    private boolean onPreparedSuccessInner() {
        logHelper.infoShow("onPreparedSuccessInner begin");
        if (!mIsPrepared) {

            try {
                initAudiotrack();
            } catch (IllegalStateException e) {
                e.printStackTrace();

                logHelper.infoShow("prepare error MEDIA_ERROR_AUDIOTRACK");
                SendErrorMessage(MEDIA_ERROR_AUDIOTRACK, 0);
                return false;
            }

            InitAdParm();
        }

        logHelper.infoShow("onPreparedSuccessInner end");
        return true;
    }

    /*
     * Do not change these values without updating their counterparts in include/media/mediaplayer.h!
     */
    private static final int MEDIA_NOP = 0; // interface test message
    private static final int MEDIA_PREPARED = 1;
    private static final int MEDIA_PLAYBACK_COMPLETE = 2;
    private static final int MEDIA_BUFFERING_UPDATE = 3;
    private static final int MEDIA_SEEK_COMPLETE = 4;
    private static final int MEDIA_ERROR = 100;
    private static final int MEDIA_INFO = 200;
    private static final int MEDIA_ADINFO_CURANCHOR_INDEX = 300;
    private static final int MEDIA_ADINFO_INSERT_END = 301;
    private static final int MEDIA_ADINFO_DOWNLOAD_ERROR = 302;
    private static final int MEDIA_ADINFO_INSERT_ERROR = 303;

    private class EventHandler extends Handler {
        private AudioPlayer mAudioPlayer;

        public EventHandler(AudioPlayer ap, Looper looper) {
            super(looper);
            mAudioPlayer = ap;
        }

        @Override
        public void handleMessage(Message msg) {
            logHelper.infoShow("handleMessage msg = " + msg.what);

            if (mAudioPlayer.mNativeContext == 0) {
                logHelper.warnShow("mediaplayer went away with unhandled events");
                return;
            }

            logHelper.infoShow("msg = " + msg.what + " begin");
            switch (msg.what) {
                case MEDIA_PREPARED:
                    if (msg.arg1 == 0) {// 0 means ok
                        // p2p 内部经过处理保证空间够用，不够时会释放空间
                        // if (isSdcardFull(native_getCurrentTaskNeedSpace() +
                        // EXTRA_SPACE_BYTE)) {
                        // logHelper.infoShow("prepare error MEDIA_ERROR_SDCARDSPACE");
                        // SendErrorMessage(MEDIA_ERROR_SDCARDSPACE, 0);
                        // } else {
                        logHelper.infoShow("msg = " + msg.what + " onPreparedSuccessInner");
                        // async. call onPreparedSuccessInner on callback
                        if (mIsPrepareAsync) {
                            if (!onPreparedSuccessInner()) {
                                break;
                            }
                        }

                        mIsPrepared = true;

                        logHelper.infoShow("msg = " + msg.what + " onPrepared");
                        if (mOnPreparedListener != null) {
                            mOnPreparedListener.onPrepared(mAudioPlayer);
                        }
                        // }
                    } else {
                        SendErrorMessage(msg.arg1, 0);
                    }

                    break;

                case MEDIA_PLAYBACK_COMPLETE:
                    if (mOnCompletionListener != null)
                        mOnCompletionListener.onCompletion(mAudioPlayer);
                    stayAwake(false);
                    break;

                case MEDIA_BUFFERING_UPDATE:
                    if (mOnBufferingUpdateListener != null)
                        mOnBufferingUpdateListener.onBufferingUpdate(mAudioPlayer, msg.arg1);
                    break;

                case MEDIA_SEEK_COMPLETE:
                    if (msg.arg1 == 0) {// 0 suc, -1 fail
                        // synchronized (AudioPlayer.this) {
                        {
                            AudioTrack auTrack = mAudioTrack;

                            try {
                                if (auTrack != null) {
                                    auTrack.flush();
                                    // mStartSamples =
                                    // audioTrack.getPlaybackHeadPosition();
                                }
                            } catch (Exception e) {
                                e.printStackTrace();
                            }
                        }

                        // seek结束后锚点无效，重新查找锚点
                        mAdCurAnchorValid = false;
//mingliang
                        logHelper.infoShow("mlcheck_MEDIA_SEEK_COMPLETE " + mStartPos);
                        mStartPosTem = mStartPos;
                        mAdTimerSizeTmp = (long) ((long) mStartPosTem * (long) mBytePerSec) / 1000;
                        mAdTimerSize = mAdTimerSizeTmp;
                    }else{
                    	mStartPos = mStartPosTem;
                    	mAudioTrackSumStart = mAudioTrackSumOriginal;          
                    }

                    // lg temp 主版暂不上报错误通知
                    if (LEARN_MODE) {
                        if (mOnSeekCompleteListener != null) {
                            if (msg.arg1 == 0) {
//                                startFadeInFadeOut(-1, SEEK_FADEIOUT_SLEEP_MS, FADE_MODE_SEEK);// fadeout
                                                                           // than
                                                                           // fadein
                                logHelper.infoShow("seek onSeekComplete ok");
                                mOnSeekCompleteListener.onSeekComplete(mAudioPlayer);
                            } else {
                                logHelper.infoShow("seek onSeekComplete error");
                                mOnSeekCompleteListener.onSeekComplete(null);
                            }
                        }
                    } else {
                        if (msg.arg1 == 0) {
                            logHelper.infoShow("seek onSeekComplete ok");
//                            startFadeInFadeOut(-1, SEEK_FADEIOUT_SLEEP_MS, FADE_MODE_SEEK);// fadeout
                                                                       // than
                                                                       // fadein
                            logHelper.infoShow("seek onSeekComplete startFadeInFadeOut ok");
                            if (mOnSeekCompleteListener != null) {
                                mOnSeekCompleteListener.onSeekComplete(mAudioPlayer);
                            }
                            logHelper.infoShow("seek startFadeInFadeOut onSeekComplete");

                        } else {
                            logHelper.infoShow("seek onSeekComplete error");
                            SendErrorMessage(MEDIA_ERROR_SEEK, 0);
                        }
                    }

                    if (msg.arg1 == 0) {
                        // insert immediately after seek mIsInsertImmediately
                        if ((mAdWillPlayValid || mIsInsertImmediately) && (mAdInfos != null) && (mAdReport != null)) {
                            logHelper.infoShow(" newmedia adplaying SingalInsert (" + mAdCurAdIndex + ","
                                    + mAdCurAnchorIndex + ")");

                            formatAdSendSignal(AD_INSERT);

                            int curanchortime = -1;
                            if ((mAdCurAnchorIndex >= 0) && (mAdCurAnchorIndex < mAdInfos.anchorNums)
                                    && (mAdInfos.anchorTime != null)) {
                                curanchortime = mAdInfos.anchorTime[mAdCurAnchorIndex];
                            }

                            logHelper.infoShow(" adseekto seekto SingalInsert (" + mStartPos + "," + curanchortime
                                    + "," + mAdCurAnchorIndex + ")");

                            mAdSingalDownloadValid = false;
                            mAdSingalInsertValid = false;

                        }
                    }
                    logHelper.infoShow("advertisement MEDIA_SEEK_COMPLETE mAdTimerMs = " + b2ms(mAdTimerSize));
                    break;
                case MEDIA_ERROR:
                    // For PV specific error values (msg.arg2) look in
                    // opencore/pvmi/pvmf/include/pvmf_return_codes.h
                    logHelper.errorShow("Error (" + msg.arg1 + "," + msg.arg2 + ")");
                    boolean error_was_handled = false;
                    if (mOnErrorListener != null) {
                        error_was_handled = mOnErrorListener.onError(mAudioPlayer, msg.arg1, msg.arg2);
                    }
                    if (mOnCompletionListener != null && !error_was_handled) {
                        mOnCompletionListener.onCompletion(mAudioPlayer);
                    }
                    stayAwake(false);
                    break;

                case MEDIA_INFO:
                    // For PV specific code values (msg.arg2) look in
                    // opencore/pvmi/pvmf/include/pvmf_return_codes.h
                    logHelper.infoShow("Info (" + msg.arg1 + "," + msg.arg2 + ")");
                    if (mOnInfoListener != null) {
                        mOnInfoListener.onInfo(mAudioPlayer, msg.arg1, msg.arg2);
                    }
                    // No real default action so far.
                    break;

                case MEDIA_NOP: // interface test message - ignore
                    break;
                case MEDIA_ADINFO_CURANCHOR_INDEX:
                    logHelper.infoShow("advertisement MEDIA_ADINFO_CURANCHOR_INDEX  = " + msg.arg1);
                    if (msg.arg1 >= 0 && (mAdInfos != null)) {
                        mAdCurAnchorIndex = msg.arg1;
                        mAdCurAnchorValid = true;
                    }
                    break;
                case MEDIA_ADINFO_INSERT_END:
                    logHelper.infoShow("MEDIA_ADINFO_INSERT_END msg = " + msg.arg1);
                    break;
                case MEDIA_ADINFO_DOWNLOAD_ERROR:
                case MEDIA_ADINFO_INSERT_ERROR:
                    if ((mAdInfos != null)) {
                        // if playing and receive error, send interrupt
                        checkAdInterrupt();
                        // insert error -1, reset status
                        resetAdStatus();
                    }
                    break;
                default:
                    logHelper.errorShow("Unknown message type " + msg.what);
                    break;
            }

            logHelper.infoShow("msg = " + msg.what + " end");
        }
    }

    private void SendErrorMessage(int arg1, int arg2) {
        if (mEventHandler != null) {
            Message m = mEventHandler.obtainMessage();
            m.what = MEDIA_ERROR;
            m.arg1 = arg1;
            m.arg2 = arg2;
            mEventHandler.sendMessage(m);
        }
    }

    /**
     * Called from native code when an interesting event happens. This method just uses the EventHandler system to post
     * the event back to the main app thread. We use a weak reference to the original AudioPlayer object so that the
     * native code is safe from the object disappearing from underneath it. (This is the cookie passed to
     * native_setup().)
     */
    private static void postEventFromNative(Object mediaplayer_ref, int what, int arg1, int arg2, Object obj) {

        logHelper.infoShow("postEventFromNative");
        AudioPlayer ap = (AudioPlayer) ((WeakReference<?>) mediaplayer_ref).get();
        if (ap == null) {
            logHelper.infoShow("postEventFromNative ap null");
            return;
        }

        if (ap.mEventHandler != null) {
            logHelper.infoShow("postEventFromNative message");
            Message m = ap.mEventHandler.obtainMessage(what, arg1, arg2, obj);
            ap.mEventHandler.sendMessage(m);
        }
    }

    private int getRenderBuffer(byte[] buffer) {
        return getRenderBuffer(buffer.length, buffer);
    }

    /*
     * inner native
     */
    protected native boolean native_advertisementSetInfos(AdvertiseMentInfos infos);

    protected native boolean native_advertisementSetAdInfos(AdvertiseMentInfos infos);

    protected native boolean native_advertisementSendSingal(AdvertiseMentInfoReport param);

    protected static native int native_getAudioDuration(String path);

    protected static native ACAudioTags native_getAudioTags(String filePath, String imagePath, String lrcPath);
   
    protected native void native_setRefrenceDuration(int durationMillisec);

    protected native int native_getDuration();

    protected native void _prepare() throws IOException, IllegalStateException;

    protected native void _prepareAsync() throws IllegalStateException;

    protected native void _start() throws IllegalStateException;

    protected native void _stop() throws IllegalStateException;

    protected native void _pause() throws IllegalStateException;

    protected native void _seekTo(int msec) throws IllegalStateException;

    protected native void _release();

    protected native void _reset();

    protected static native final void native_init();

    protected native final void native_setup(Object mediaplayer_this);

    protected native final void native_finalize();

    private native int getRenderBuffer(int bsize, byte[] buffer);

    private native int getRenderSampleRate();

    private native int getRenderChannels();

    private native int getRenderBitsPerSample();

    private native void _getFreqData(int nPlayPos, int datacount, int[] data);

    private native void _setVisualBuffer(int buffersize);

    protected native void setEffectParam(int nEffectIndex, int nParamIndex, short nIndex, short nData);

    protected native short getEffectParam(int nEffectIndex, int nParamIndex, short nIndex);

    protected native void getEffectParamRange(int nEffectIndex, int nParamIndex, int[] data);

    protected native void native_setAudioEffectCombine(String aebFilePath) throws IllegalStateException, IllegalArgumentException, RuntimeException;
    
    protected static native void native_getPresetEQLevels(short preset, int[] levels);

    // protected native boolean _isPlaying();

    private native static String native_getCrashTraceInfo() throws IllegalStateException;

    /*
     * 
     * private static final int SingalUnknown = 0x00; private static final int SingalDownload = 0x01; private static
     * final int SingalInsert = 0x02; private static final int SingalPlaying = 0x04; private static final int
     * SingalFinished = 0x08;
     * 
     * 
     * Message msg = advertiseMentHandler.obtainMessage(); msg.what = SingalDownload;
     * advertiseMentHandler.sendMessage(msg);
     */
    // ///----advertisement---------//

    private void advertiseMentSendMessage(int msg, int adindex) {
        Message admsg = advertiseMentHandler.obtainMessage();
        admsg.what = msg;
        admsg.arg1 = adindex;
        advertiseMentHandler.sendMessage(admsg);
    }

    // report audiocore;
    private void formatAdSendSignal(int signal) {
        if (null == mAdReport) {
            return;
        }

        mAdReport.singal = signal;
        mAdReport.curIndex = mAdCurAdIndex;
        mAdReport.timerAbsoluteSumMs = b2ms(mAdTimerSize);
        mAdReport.timerRelativeSumMs = b2ms(mAdTimerRelativeSize);
        // report audiocore;
        advertisementSendSingal(mAdReport);

    }

    private Handler advertiseMentHandler = new Handler(new Handler.Callback() {
        @Override
        public boolean handleMessage(Message msg) {
            logHelper.infoShow("policy adsmsg = " + msg.what + " msg.arg1 " + msg.arg1);
            if (mAdReport == null || (msg.what > AD_MAX)) {
                logHelper.infoShow("adsmsg error = " + msg.what);
                return false;
            }

            // report audiocore;
            formatAdSendSignal(msg.what);

            // report app;
            if ((mOnAdvertiseMentPlayStatusListener != null)) {
                if ((mAdInfos != null) && (mAdInfos.anchorNums > 0) && (mAdInfos.adNums > 0) && (mAdInfos.adId != null)) {
                    int adindex = msg.arg1;
                    int adid = -1;
                    if (adindex >= 0 && adindex < mAdInfos.adNums) {
                        adid = mAdInfos.adId[adindex];
                        logHelper.infoShow("policy mAdReport.singal = " + mAdReport.singal + " adindex " + adindex
                                + " adid " + adid);
                    }

                    mOnAdvertiseMentPlayStatusListener.onAdvertiseMentPlayStatus(AudioPlayer.this, mAdReport.singal,
                            adid, 0);
                } else {
                    mOnAdvertiseMentPlayStatusListener.onAdvertiseMentPlayStatus(AudioPlayer.this, mAdReport.singal,
                            -1, 0);
                }
            }
            return true;
        }
    });

    private static final int FADE_SLEEP_MS = 50;
    private static final int FADEIN_MUTE_MS = 0;
    private static final int FADEIN_SLEEP_MS = 1000;
    private static final int SEEK_FADEIOUT_SLEEP_MS = 400;
    private static final int PAUSE_FADEIOUT_SLEEP_MS = 400;
    
    private static final int FADE_MODE_PLAY = 0;
    private static final int FADE_MODE_PAUSE = 1;
    private static final int FADE_MODE_SEEK = 2;
    private static final int FADE_MODE_STOP = 3;

    private int mFadeReferentCount = 0;

    // -1 means default value, 0 means disable
    private void startFadeInFadeOut(final int nFadeIn, final int nFadeOut, final int nFadeMode) {
        if (!mIsEnableFadeInFadeOut) {
            return;
        }

        logHelper.infoShow("startFadeInFadeOut nFadeIn " + nFadeIn + " nFadeOut " + nFadeOut + " nFadeMode "
                + nFadeMode + " mFadeReferentCount " + mFadeReferentCount);

        final AudioTrack audioTrack = mAudioTrack;
        if (audioTrack == null) {
            return;
        }
        // may be block a very long time by audiotrack in renderthread
        // synchronized(this) {
        // when stop, use audioTrack, mAudioTrack set null, create new
        // mAudioTrack next play. no lock
        if (FADE_MODE_STOP == nFadeMode) {
            mAudioTrack = null;
        }
        // }

        logHelper.infoShow("startFadeInFadeOut Thread new begin");

        if ((FADE_MODE_PLAY == nFadeMode) && (nFadeIn != 0) && (0 == mFadeReferentCount)) {
        	logHelper.infoShow("startFadeInFadeOut FADE_MODE_PLAY set silence when begin");
        	setVolumeSys(0.0f, 0.0f, audioTrack);
        }
        
        new Thread("FadeInFadeOut") {
            public void run() {
                // stoped state not fade
                logHelper.infoShow("startFadeInFadeOut Thread run begin" + " nFadeMode " + nFadeMode);

                int nPlayState = AudioTrack.PLAYSTATE_STOPPED;
                try {
                    nPlayState = audioTrack.getPlayState();
                } catch (Exception e) {
                    e.printStackTrace();
                }

                // need reset volume before return
                if (AudioTrack.PLAYSTATE_PLAYING != nPlayState && AudioTrack.PLAYSTATE_PAUSED != nPlayState) {
                	setVolumeSys(mVolumeLeft, mVolumeRight, audioTrack);
                	logHelper.infoShow("startFadeInFadeOut Thread run exit, nPlayState is " + nPlayState + " nFadeMode " + nFadeMode);
                    return;
                }
                if (mBytePerSec == 0) {
                    return;
                }

                do {
                    logHelper.infoShow("FadeInFadeOut enter mFadeReferentCount " + mFadeReferentCount + " nFadeMode "
                            + nFadeMode);
                    ++mFadeReferentCount;

                    int nFadeMs = 0;
                    int nCurMs = nFadeMs;
                    // fade out
                    if (nFadeOut != 0) {
                        logHelper.infoShow("startFadeInFadeOut Thread run fadeout begin" + " nFadeMode " + nFadeMode);

                        nFadeMs = (nFadeOut == -1) ? (int) (1000.0f * mBufferSize / mBytePerSec) : nFadeOut;
                        if (nFadeMs == 0) {
                            --mFadeReferentCount;
                            break;
                        }
                        nCurMs = nFadeMs;

                        while (true) {
                            nCurMs -= FADE_SLEEP_MS;
                            float fVol = nCurMs / (float) nFadeMs;
                            if (fVol < 0.0f) {
                                fVol = 0.0f;
                            }

                            if (!mIsPrepared) {
                            	logHelper.infoShow("startFadeInFadeOut Thread need exit fadein" + " nFadeMode " + nFadeMode);
                            	return;
                            }
                            
                            setVolumeSys(mVolumeLeft * fVol, mVolumeRight * fVol, audioTrack);
                            try {
                                Thread.sleep(FADE_SLEEP_MS);
                            } catch (InterruptedException e) {
                                // TODO Auto-generated catch block
                                e.printStackTrace();
                            }

                            if (nCurMs < -FADE_SLEEP_MS) {
                                break;
                            }
                        }

                        if ((FADE_MODE_PAUSE == nFadeMode) && (1 == mFadeReferentCount)) {
                            logHelper.infoShow("FadeInFadeOut pause");

                            try {
                                if (audioTrack != null) {
                                    audioTrack.pause();
                                }
                            } catch (Exception e) {
                                e.printStackTrace();
                            }
                        }
                        // if need stop, release current audiotrack
                        if (FADE_MODE_STOP == nFadeMode) {
                            logHelper.infoShow("FadeInFadeOut stop" + " nFadeMode " + nFadeMode);
                            synchronized (this) {
                                try {
                                    if (audioTrack != null) {
                                        audioTrack.pause();
                                        audioTrack.flush();
                                        audioTrack.stop();
                                        audioTrack.release();
                                    }
                                } catch (Exception e) {
                                    e.printStackTrace();
                                }
                            }
                        }

                        // cost older buffer of audiotrack,flush is useless
                        if (FADE_MODE_SEEK == nFadeMode) {
                        	for (int i=0; i<10; ++i) {
                                if (!mIsPrepared) {
                                	logHelper.infoShow("startFadeInFadeOut Thread need exit fadeout silence" + " nFadeMode " + nFadeMode);
                                	return;
                                }

                                try {
                                    Thread.sleep(10);
                                } catch (InterruptedException e) {
                                    // TODO Auto-generated catch block
                                    e.printStackTrace();
                                }
                            }
                        }
                        
                        logHelper.infoShow("startFadeInFadeOut Thread run fadeout end" + " nFadeMode " + nFadeMode);
                    }
                    
                    // fade in
                    if (nFadeIn != 0) {
                        logHelper.infoShow("startFadeInFadeOut Thread run fadein begin" + " nFadeMode " + nFadeMode);
                        nFadeMs = (nFadeIn == -1) ? FADEIN_SLEEP_MS : nFadeIn;
                        nCurMs = 0;
                        if (nFadeMs == 0) {
                            --mFadeReferentCount;
                            break;
                        }
                    	
                    	float fVol = 0.0f;
                        while (true) {
                            fVol = nCurMs / (float) nFadeMs;
                            setVolumeSys(mVolumeLeft * fVol, mVolumeRight * fVol, audioTrack);

                            nCurMs += FADE_SLEEP_MS;
                            
                            if (!mIsPrepared) {
                            	logHelper.infoShow("startFadeInFadeOut Thread need exit fadein" + " nFadeMode " + nFadeMode);
                            	return;
                            }
                            
                            try {
                                Thread.sleep(FADE_SLEEP_MS);
                            } catch (InterruptedException e) {
                                // TODO Auto-generated catch block
                                e.printStackTrace();
                            }

                            if (nCurMs > nFadeMs) {
                                break;
                            }
                        }

                        if(fVol < 1.0f) {
                            // set volume to max at last
                            setVolumeSys(mVolumeLeft, mVolumeRight, audioTrack);
                        }

                        logHelper.infoShow("startFadeInFadeOut Thread run fadein end" + " nFadeMode " + nFadeMode);
                    }

                    --mFadeReferentCount;
                    logHelper.infoShow("FadeInFadeOut exit mFadeReferentCount " + mFadeReferentCount + " nFadeMode "
                            + nFadeMode);
                } while (false);
                
                if(mFadeReferentCount < 0) {
                	logHelper.infoShow("startFadeInFadeOut Thread run mFadeReferentCount error, need normal " + " nFadeMode " + nFadeMode + " mFadeReferentCount " + mFadeReferentCount);
                	mFadeReferentCount = 0;
                }
                
                logHelper.infoShow("startFadeInFadeOut Thread run end " + " nFadeMode " + nFadeMode);
            }
        }.start();

        logHelper.infoShow("startFadeInFadeOut Thread new end");
    }

    private void setVolumeSys(float leftVolume, float rightVolume, AudioTrack refAudioTrack) {

        // add audiotrack setVolume
        // logHelper.infoShow("setVolume left:" + leftVolume + "right:" +
        // rightVolume);
        AudioTrack audioTrack = (refAudioTrack == null) ? mAudioTrack : refAudioTrack;
        if (audioTrack != null) {
            float fRange = AudioTrack.getMaxVolume() - AudioTrack.getMinVolume();
            if (fRange > 0.0f) {
                leftVolume = leftVolume / fRange + AudioTrack.getMinVolume();
                rightVolume = rightVolume / fRange + AudioTrack.getMinVolume();
            } else {
                logHelper.errorShow("setVolume getVoumeMaxMin error");
            }
            try {
                audioTrack.setStereoVolume(leftVolume, rightVolume);
            } catch (IllegalStateException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        } else {
            logHelper.errorShow("setVolume mAudioTrack = null");
        }
    }

    /**
     * 获得sd卡剩余容量，即可用大小
     * 
     * @return
     */
    private static long getSDAvailableSize() {
        long size = 0L;
        try {
            File path = Environment.getExternalStorageDirectory();
            StatFs stat = new StatFs(path.getPath());
            long blockSize = stat.getBlockSize();
            long availableBlocks = stat.getAvailableBlocks();
            size = blockSize * availableBlocks;
            logHelper.errorShow("getSDAvailableSize path " + path.getPath() + " size " + size);
        } catch (Exception e) {
            e.printStackTrace();
            size = 0L;
        }
        return size;
    }

    /**
     * Check if we can allocate some bytes in storage.
     * 
     * @param sizeToAllocateInBytes
     * @return
     */
    private static boolean isSdcardFull(long sizeToAllocateInBytes) {
        long availableSpare = getSDAvailableSize();
        logHelper.errorShow("isSdcardFull availableSpare " + availableSpare + " sizeToAllocateInBytes "
                + sizeToAllocateInBytes);
        return (availableSpare < sizeToAllocateInBytes);
    }

    // string length must less than 8
    private static String getProductId(int pId) {
        String strProduct = new String();
        int product = pId & 0xFF000000;
        switch (product) {
            case PID_MAIN_MUSIC:
                strProduct = "Android";
                break;
            case PID_MAIN_LEBO:
                strProduct = "ALebo";
                break;
            case PID_SDK_MUSIC:
                strProduct = "ASDK";
                break;
            case PID_SDK_LEBO:
                strProduct = "ASDKLeb";
                break;
            case PID_SHOUBAI_MUSIC:
                strProduct = "ABaidu";
                break;
            case PID_YYR:
                strProduct = "AYYR";
                break;
            case PID_LAVA:
                strProduct = "ALava";
                break;
            default:
                strProduct = "unkown";
                break;
        }

        return strProduct;
    }

    /**
     * get environment info from context and set to audiocore.
     * 
     * @param pId : id of product
     * @return
     */
    private static void setEnvironmentInfo(int pId) {
        if (longLifeContext != null) {
        	if (hasReadPhoneStatePermission()) {
                final TelephonyManager tm = (TelephonyManager) longLifeContext.getSystemService(Context.TELEPHONY_SERVICE);
                final String tmDevice, tmSerial, androidId;
                tmDevice = "" + tm.getDeviceId();
                tmSerial = "" + tm.getSimSerialNumber();
                androidId =
                        ""
                                + android.provider.Settings.Secure.getString(longLifeContext.getContentResolver(),
                                        android.provider.Settings.Secure.ANDROID_ID);
                UUID deviceUuid =
                        new UUID((androidId + tmDevice).hashCode(), ((long) (tmDevice + tmSerial).hashCode() << 32)
                                | (tmSerial + androidId).hashCode());
                String ci = deviceUuid.toString();
                String machine = android.os.Build.MODEL;
                String sys = getProductId(pId);
                String sysver = android.os.Build.VERSION.RELEASE;
                String appname = longLifeContext.getPackageName();
                String appver = "";
                try {
                    PackageManager packageManager = longLifeContext.getPackageManager();
                    PackageInfo packageInfo = packageManager.getPackageInfo(longLifeContext.getPackageName(), 0);
                    appver = packageInfo.versionName;
                    if (appver == null) {
                        appver = "";
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                }
                logHelper.debugShow("SetEnvironmentInfo:" + ci + "," + machine + "" + "," + sys + "," + sysver + ","
                        + appname + "," + appver + ",");
                setEnvironmentInfo(ci, machine, sys, sysver, appname, appver);
        	} else {
        		logHelper.debugShow("do not have read state permission");
        		String sys = getProductId(pId);
        		setEnvironmentInfo("", "", sys, "", "", "");
        	}
        }
    }

    private static final String READ_PHONE_STATE_PERMISSION = "android.permission.READ_PHONE_STATE";
    private static boolean hasReadPhoneStatePermission() {
        int perm = longLifeContext.checkCallingOrSelfPermission(READ_PHONE_STATE_PERMISSION);
        return perm == PackageManager.PERMISSION_GRANTED;
    }

    // ////////////////////////////////////////////////////////////////////////////////////
    // p2p
    
    private static native void native_setProxy(String host, int port, String user, String passwd, int authType);

    private static native void setEnvironmentInfo(String ci, String machine, String sys, String sysver, String appname,
            String appver);

    private static void onHttpResponseNotify(String url, int code, String text) {        
        if (mOnHttpResponseListener != null) {
            logHelper.debugShow("onTaskStateNotify url " + url + 
                    " code " + code + " text " + text);
            mOnHttpResponseListener.onHttpResponse(url, code, text);
            mOnHttpResponseListener = null;
        }

    }

    private static final int STATE_CHANGED = 0;
    private static final int NOTIFY_ERROR = 1;
    private static final int NOTIFY_PROGRESS = 2;
    private static final int NOTIFY_TIMEOUT = 3;
    private static final int NOTIFY_DOWNLOADBYTES = 4;

    private static void onTaskStateNotify(int nInfo, int[] nParams, String url, String ip) {
        try {
            onTaskStateNotifyInner(nInfo, nParams, url, ip);
        }    catch (Exception ex) {
            ex.printStackTrace();
        }
    }
    
    private static void onTaskStateNotifyInner(int nInfo, int[] nParams, String url, String ip) {
        synchronized (lock) {
            if (null == nParams) {
                return;
            }

            if (NOTIFY_DOWNLOADBYTES == nInfo) {
                if (mOnDownloadBytesListener != null) {
                    mOnDownloadBytesListener.onDownloadBytes(nParams[1] + nParams[2]);
                }
            } else {
                String strTaskIdString = String.valueOf(nParams[0]);
                if (null == taskDownloadDelegates) {
                    return;
                }
                if (taskDownloadDelegates.containsKey(strTaskIdString)) {
                    if (null == downloadTaskHandler) {
                        return;
                    }
                    
                    Message msg = downloadTaskHandler.obtainMessage();
                    Bundle bundle = new Bundle();
                    bundle.putIntArray("param", nParams);
                    msg.what = nInfo;
                    msg.setData(bundle);
                    downloadTaskHandler.sendMessage(msg);
                }
            }
        }
    }

    private static Handler downloadTaskHandler = null;

    private static void CreateDownloadTaskHandler() {
        if (downloadTaskHandler != null) {
            return;
        }

        downloadTaskHandler = new Handler(new Handler.Callback() {

            @Override
            public boolean handleMessage(Message msg) {
                // TODO Auto-generated method stub
                Bundle bundle = null;
                int[] nParams = null;
                String strTaskId = null;
                OnTaskDownloadDelegate delegate = null;
                switch (msg.what) {
                    case STATE_CHANGED:
                        bundle = msg.getData();
                        nParams = bundle.getIntArray("param");
                        strTaskId = String.valueOf(nParams[0]);
                        if (taskDownloadDelegates.containsKey(strTaskId)) {
                            delegate = taskDownloadDelegates.get(strTaskId);
                            if (delegate != null) {
                                delegate.onTaskIdStatusChanged(nParams[0], nParams[1], nParams[2]);
                            } else {
                                taskDownloadDelegates.remove(strTaskId);
                            }

                        }
                        break;
                    case NOTIFY_ERROR:
                        bundle = msg.getData();
                        nParams = bundle.getIntArray("param");
                        strTaskId = String.valueOf(nParams[0]);
                        if (taskDownloadDelegates.containsKey(strTaskId)) {
                            delegate = taskDownloadDelegates.get(strTaskId);
                            if (delegate != null) {
                                logHelper.debugShow("downloadtask NOTIFY_ERROR taskID:" + nParams[0] + " state : "
                                        + nParams[1] + " errorcode: " + nParams[2]);
                                delegate.onTaskIdErrorCode(nParams[0], nParams[1], nParams[2]);
                            } else {
                                taskDownloadDelegates.remove(strTaskId);
                            }
                        }
                        break;
                    case NOTIFY_PROGRESS:
                        bundle = msg.getData();
                        nParams = bundle.getIntArray("param");
                        strTaskId = String.valueOf(nParams[0]);
                        if (taskDownloadDelegates.containsKey(strTaskId)) {
                            delegate = taskDownloadDelegates.get(strTaskId);
                            if (delegate != null) {
                                delegate.onTaskIdProgress(nParams[0], nParams[1]);
                            } else {
                                taskDownloadDelegates.remove(strTaskId);
                            }
                        }
                        break;
                    case NOTIFY_TIMEOUT:
                        bundle = msg.getData();
                        nParams = bundle.getIntArray("param");
                        strTaskId = String.valueOf(nParams[0]);
                        if (taskDownloadDelegates.containsKey(strTaskId)) {
                            delegate = taskDownloadDelegates.get(strTaskId);
                            if (delegate != null) {
                                delegate.onTaskIdTimeout(nParams[0]);
                            } else {
                                taskDownloadDelegates.remove(strTaskId);
                            }
                        }
                        break;
                    default:
                        break;
                }
                return false;
            }
        });
    }

    private static int postTask(int nType, String url, OnTaskDownloadDelegate delegate) {
        int nTaskId = 0;
        if (url != null && url.length() > 0) {
            nTaskId = createDownloadTask(nType, url);
            if (nTaskId > 0) {
                String strTaskIdString = String.valueOf(nTaskId);
                if (delegate != null && !taskDownloadDelegates.containsKey(strTaskIdString)) {
                    taskDownloadDelegates.put(strTaskIdString, delegate);
                }
            }
        }
        return nTaskId;
    }

    private static native int createDownloadTask(int nType, String url);

    private static native int deleteTask(int nTaskId);

    private static native void relationTasks(int fatherTaskId, int sonTaskId);

    private static void onFeedbackNotify(String feedback) {
        logHelper.debugShow("Feedback:" + feedback);
    }

    private native int native_getCurrentTaskNeedSpace();

    private static native void native_setUserAgent(String ua) throws IllegalArgumentException;

    private static native void native_setPartner(String partner) throws IllegalArgumentException;

    private static native void native_setCustomHttpHeader(
            String keys, String values, String seperator) throws IllegalArgumentException;
    
    private native String native_getCurFilePathString() throws IllegalStateException;

    private static native void native_setNativeCrashFilePath(String strFilePath) throws IllegalArgumentException;

    // 设置上传下载类型
    private static native void native_setUpDownloadType(int type);

    private static native void native_setDownloadTaskTimeout(int timeoutsecond);

    private native void native_setDataSource(String path) throws IOException, IllegalArgumentException,
            IllegalStateException;

    // stop current download task
    private native void native_stopCurrentTask() throws IllegalStateException;

    private static native void native_enableDownloadBytesCallback();

    private static native void native_httpRequest(String url, 
            int retrytime, int timeout, String headers[]) throws IllegalArgumentException;

    private static native boolean native_isUrlCached(String httpurl, String hash);

    private static void setUpDownloadTypeByNetType() {
        logHelper.infoShow("setUpDownloadTypeByNetType : " + mNetState);
        switch (mNetState) {
            case NETWORK_TYPE_WIFI:
                native_setUpDownloadType(NETWORK_HTTP_DOWNLOAD | NETWORK_P2P_DOWNLOAD);
                break;
            case NETWORK_TYPE_2G:
                native_setUpDownloadType(NETWORK_HTTP_DOWNLOAD);
                break;
            case NETWORK_TYPE_3G:
            case NETWORK_TYPE_4G:
                if (mEnableMobileDataP2P) {
                    native_setUpDownloadType(NETWORK_HTTP_DOWNLOAD | NETWORK_P2P_DOWNLOAD);
                } else {
                    native_setUpDownloadType(NETWORK_HTTP_DOWNLOAD);
                }
                break;
            default:
                native_setUpDownloadType(NETWORK_HTTP_DOWNLOAD);
                break;
        }
    }
}
